#!/usr/bin/env python3

from trojanvision.configs import config
from trojanvision.datasets import ImageSet
from trojanzoo.environ import env
from trojanzoo.utils.module import BasicObject
from trojanzoo.utils.output import ansi

import torch
import torchvision.transforms.functional as F

import numpy as np
import os
import PIL.Image as Image

import argparse
from trojanzoo.configs import Config
from collections.abc import Callable

dir_path = os.path.dirname(__file__)


def get_edge_color(
    mark: torch.Tensor,
    mark_background_color: str | torch.Tensor = 'auto'
) -> torch.Tensor | None:
    # if any pixel is not fully opaque
    if not mark[-1].allclose(torch.ones_like(mark[-1]), atol=1e-3):
        return None
    mark = mark[:-1]    # remove alpha channel
    match mark_background_color:
        case torch.Tensor():
            return torch.as_tensor(mark_background_color).expand(mark.size(0))
        case 'black':
            return torch.zeros(mark.size(0))
        case 'white':
            return torch.ones(mark.size(0))
        case 'auto':
            if mark.flatten(1).std(1).max() < 1e-3:
                return None
            else:
                _list = [mark[:, 0, :], mark[:, -1, :],
                         mark[:, :, 0], mark[:, :, -1]]
                return torch.cat(_list, dim=1).mode(dim=-1)[0]
        case _:
            raise ValueError(f'{mark_background_color=:s}')


def update_mark_alpha_channel(
    mark: torch.Tensor,
    mark_background_color: torch.Tensor | None = None
) -> torch.Tensor:
    if mark_background_color is None:
        return mark
    mark = mark.clone()
    mark_background_color = mark_background_color.view(-1, 1, 1)
    mark[-1] = ~mark[:-1].isclose(mark_background_color,
                                  atol=1e-3).all(dim=0)
    return mark


class Watermark(BasicObject):
    r"""Watermark class that is used for backdoor attacks.

    Note:
        Images with alpha channel are supported.
        In this case, :attr:`mark_alpha` will be multiplied.

    Warning:
        :attr:`mark_random_init` and :attr:`mark_scattered` can't be used together.

    Args:
        mark_path (str):
            | Path to watermark image or npy file.
              There are some preset marks in the package.
            | Defaults to ``'square_white.png'``.

            .. table::
                :widths: auto

                +-----------------------------+---------------------+
                |      mark_path              |    mark image       |
                +=============================+=====================+
                |  ``'apple_black.png'``      |  |apple_black|      |
                +-----------------------------+---------------------+
                |  ``'apple_white.png'``      |  |apple_white|      |
                +-----------------------------+---------------------+
                |  ``'square_black.png'``     |  |square_black|     |
                +-----------------------------+---------------------+
                |  ``'square_white.png'``     |  |square_white|     |
                +-----------------------------+---------------------+
                |  ``'watermark_black.png'``  |  |watermark_black|  |
                +-----------------------------+---------------------+
                |  ``'watermark_white.png'``  |  |watermark_white|  |
                +-----------------------------+---------------------+
        data_shape (list[int]): The shape of image data ``[C, H, W]``.

            See Also:
                Usually passed by ``dataset.data_shape``.
                See :attr:`data_shape` from
                :class:`trojanvision.datasets.ImageSet`.
        mark_background_color (str | torch.Tensor): Mark background color.
            If :class:`str`, choose from ``['auto', 'black', 'white']``;
            else, it shall be 1-dim tensor ranging in ``[0, 1]``.
            It's ignored when alpha channel in watermark image.
            Defaults to ``'auto'``.
        mark_alpha (float): Mark opacity. Defaults to ``1.0``.
        mark_height (int): Mark resize height. Defaults to ``3``.
        mark_width (int): Mark resize width. Defaults to ``3``.

            Note:
                :attr:`self.mark_height` and :attr:`self.mark_width` will be different
                from the passed argument values
                when :attr:`mark_scattered` is ``True``.
        mark_height_offset (int): Mark height offset. Defaults to ``0``.
        mark_width_offset (int): Mark width offset. Defaults to ``0``.

            Note:
                :attr:`mark_height_offset` and
                :attr:`mark_width_offset` will be ignored
                when :attr:`mark_random_pos` is ``True``.
        mark_random_init (bool): Whether to randomly set pixel values of watermark,
            which means only using the mark shape from the watermark image.
            Defaults to ``False``.
        mark_random_pos (bool): Whether to add mark at random location when calling :meth:`add_mark()`.
            If ``True``, :attr:`mark_height_offset` and :attr:`mark_height_offset` will be ignored.
            Defaults to ``False``.
        mark_scattered (bool): Random scatter mark pixels
            in the entire image to get the watermark. Defaults to ``False``.
        mark_scattered_height (int | None): Scattered mark height. Defaults to data_shape[1].
        mark_scattered_width (int | None): Scattered mark width. Defaults to data_shape[2].

            Note:
                - The random scatter process only occurs once at watermark initialization.
                  :meth:`add_mark()` will still add the same scattered mark to images.
                - Mark image will first resize to ``(mark_height, mark_width)`` and then
                  scattered to ``(mark_scattered_height, mark_scattered_width)``.
                  If they are the same, it's actually pixel shuffling.
                - :attr:`self.mark_height` and :attr:`self.mark_width` will be set to scattered version.
        add_mark_fn (~collections.abc.Callable | None):
            Customized function to add mark to images for :meth:`add_mark()` to call.
            ``add_mark_fn(_input, mark_random_pos=mark_random_pos, mark_alpha=mark_alpha, **kwargs)``
            Defaults to ``None``.

    Attributes:
        mark (torch.Tensor): Mark float tensor with shape
            ``(data_shape[0] + 1, mark_height, mark_width)``
            (last dimension is alpha channel).
        mark_alpha (float): Mark opacity. Defaults to ``1.0``.
        mark_height (int): Mark resize height. Defaults to ``3``.
        mark_width (int): Mark resize width. Defaults to ``3``.

            Note:
                :attr:`self.mark_height` and :attr:`self.mark_width` will be different
                from the passed argument values
                when :attr:`mark_scattered` is ``True``.
        mark_height_offset (int): Mark height offset. Defaults to ``0``.
        mark_width_offset (int): Mark width offset. Defaults to ``0``.

            Note:
                :attr:`mark_height_offset` and
                :attr:`mark_width_offset` will be ignored
                when :attr:`mark_random_pos` is ``True``.
        mark_random_init (bool): Whether to randomly set pixel values of watermark,
            which means only using the mark shape from the watermark image.
            Defaults to ``False``.
        mark_random_pos (bool): Whether to add mark at random location when calling :meth:`add_mark()`.
            If ``True``, :attr:`mark_height_offset` and :attr:`mark_height_offset` will be ignored.
            Defaults to ``False``.
        mark_scattered (bool): Random scatter mark pixels
            in the entire image to get the watermark. Defaults to ``False``.
        mark_scattered_height (int): Scattered mark height. Defaults to data_shape[1].
        mark_scattered_width (int): Scattered mark width. Defaults to data_shape[2].
        add_mark_fn (~collections.abc.Callable | None):
            Customized function to add mark to images for :meth:`add_mark()` to call.
            ``add_mark_fn(_input, mark_random_pos=mark_random_pos, mark_alpha=mark_alpha, **kwargs)``
            Defaults to ``None``.

    .. |apple_black| image:: ../../../trojanvision/marks/apple_black.png
        :height: 50px
        :width: 50px
    .. |apple_white| image:: ../../../trojanvision/marks/apple_white.png
        :height: 50px
        :width: 50px
    .. |square_black| image:: ../../../trojanvision/marks/square_black.png
        :height: 50px
        :width: 50px
    .. |square_white| image:: ../../../trojanvision/marks/square_white.png
        :height: 50px
        :width: 50px
    .. |watermark_black| image:: ../../../trojanvision/marks/watermark_black.png
        :height: 50px
        :width: 50px
    .. |watermark_white| image:: ../../../trojanvision/marks/watermark_white.png
        :height: 50px
        :width: 50px
    """
    name: str = 'mark'

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup):
        r"""Add watermark arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('--mark_background_color',
                           choices=['auto', 'black', 'white'],
                           help='background color in watermark image. '
                           'It\'s ignored when alpha channel is in watermark image. '
                           '(default: "auto")')
        group.add_argument('--mark_path', help='watermark path (image or npy file), '
                           'default: "square_white.png")')
        group.add_argument('--mark_alpha', type=float,
                           help='mark opacity (default: 1.0)')
        group.add_argument('--mark_height', type=int,
                           help='mark height (default: 3)')
        group.add_argument('--mark_width', type=int,
                           help='mark width (default: 3)')
        group.add_argument('--mark_height_offset', type=int,
                           help='mark height offset (default: 0)')
        group.add_argument('--mark_width_offset', type=int,
                           help='mark width offset (default: 0)')
        group.add_argument('--mark_random_pos', action='store_true',
                           help='Random offset Location for add_mark.')
        group.add_argument('--mark_random_init', action='store_true',
                           help='random values for mark pixel.')
        group.add_argument('--mark_scattered',
                           action='store_true', help='Random scatter mark pixels.')
        group.add_argument('--mark_scattered_height', type=int,
                           help='Scattered mark height (default: same as input image)')
        group.add_argument('--mark_scattered_width', type=int,
                           help='Scattered mark width (default: same as input image)')
        return group

    def __init__(self, mark_path: str = 'square_white.png',
                 data_shape: list[int] = None, mark_background_color: str | torch.Tensor = 'auto',
                 mark_alpha: float = 1.0, mark_height: int = 3, mark_width: int = 3,
                 mark_height_offset: int = 0, mark_width_offset: int = 0,
                 mark_random_init: bool = False, mark_random_pos: bool = False,
                 mark_scattered: bool = False,
                 mark_scattered_height: int = None,
                 mark_scattered_width: int = None,
                 add_mark_fn: Callable[..., torch.Tensor] = None, **kwargs):
        super().__init__(**kwargs)
        self.param_list: dict[str, list[str]] = {}
        self.param_list['mark'] = ['mark_path',
                                   'mark_alpha', 'mark_height', 'mark_width',
                                   'mark_random_init', 'mark_random_pos',
                                   'mark_scattered']
        if not mark_random_pos:
            self.param_list['mark'].extend(['mark_height_offset', 'mark_width_offset'])
        assert mark_height > 0 and mark_width > 0
        # --------------------------------------------------- #

        self.mark_alpha = mark_alpha
        self.mark_path = mark_path
        self.mark_height = mark_height
        self.mark_width = mark_width
        self.mark_height_offset = mark_height_offset
        self.mark_width_offset = mark_width_offset
        self.mark_random_init = mark_random_init
        self.mark_random_pos = mark_random_pos
        self.mark_scattered = mark_scattered
        self.mark_scattered_height = mark_scattered_height or data_shape[1]
        self.mark_scattered_width = mark_scattered_width or data_shape[2]
        self.add_mark_fn = add_mark_fn
        self.data_shape = data_shape
        # --------------------------------------------------- #

        self.mark = self.load_mark(mark_img=mark_path,
                                   mark_background_color=mark_background_color)

    def add_mark(self, _input: torch.Tensor, mark_random_pos: bool = None,
                 mark_alpha: float = None, mark: torch.Tensor = None,
                 **kwargs) -> torch.Tensor:
        r"""Main method to add watermark to a batched input image tensor ranging in ``[0, 1]``.

        Call :attr:`self.add_mark_fn()` instead if it's not ``None``.

        Args:
            _input (torch.Tensor): Batched input tensor
                ranging in ``[0, 1]`` with shape ``(N, C, H, W)``.
            mark_random_pos (bool | None): Whether to add mark at random location.
                Defaults to :attr:`self.mark_random_pos`.
            mark_alpha (float | None): Mark opacity. Defaults to :attr:`self.mark_alpha`.
            mark (torch.Tensor | None): Mark tensor. Defaults to :attr:`self.mark`.
            **kwargs: Keyword arguments passed to `self.add_mark_fn()`.
        """
        mark_alpha = mark_alpha if mark_alpha is not None else self.mark_alpha
        mark = mark if mark is not None else self.mark
        mark_random_pos = mark_random_pos if mark_random_pos is not None else self.mark_random_pos
        if callable(self.add_mark_fn):
            return self.add_mark_fn(_input, mark_random_pos=mark_random_pos,
                                    mark_alpha=mark_alpha, **kwargs)
        trigger_input = _input.clone()
        mark = mark.clone().to(device=_input.device)

        mark_rgb_channel = mark[..., :-1, :, :]
        mark_alpha_channel = mark[..., -1, :, :].unsqueeze(-3)
        mark_alpha_channel *= mark_alpha
        if mark_random_pos:
            batch_size = _input.size(0)
            h_start = torch.randint(high=_input.size(-2) - self.mark_height, size=[batch_size])
            w_start = torch.randint(high=_input.size(-1) - self.mark_width, size=[batch_size])
            h_end, w_end = h_start + self.mark_height, w_start + self.mark_width
            for i in range(len(_input)):    # TODO: any parallel approach?
                org_patch = _input[i, :, h_start[i]:h_end[i], w_start[i]:w_end[i]]
                trigger_patch = org_patch + mark_alpha_channel * (mark_rgb_channel - org_patch)
                trigger_input[i, :, h_start[i]:h_end[i], w_start[i]:w_end[i]] = trigger_patch
                return trigger_input
        h_start, w_start = self.mark_height_offset, self.mark_width_offset
        h_end, w_end = h_start + self.mark_height, w_start + self.mark_width
        org_patch = _input[..., h_start:h_end, w_start:w_end]
        trigger_patch = org_patch + mark_alpha_channel * (mark_rgb_channel - org_patch)
        trigger_input[..., h_start:h_end, w_start:w_end] = trigger_patch

        return trigger_input

    def get_mask(self) -> torch.Tensor:
        mask = torch.zeros(self.data_shape[-2:], device=self.mark.device)
        h_start, w_start = self.mark_height_offset, self.mark_width_offset
        h_end, w_end = h_start + self.mark_height, w_start + self.mark_width
        mask[h_start:h_end, w_start:w_end].copy_(self.mark[-1])
        return mask

    @staticmethod
    def scatter_mark(mark_unscattered: torch.Tensor,
                     mark_scattered_shape: list[int]) -> torch.Tensor:
        r"""Scatter the original mark tensor to a provided shape.

        If the shape are the same, it becomes a pixel shuffling process.

        Args:
            mark_unscattered (torch.Tensor): The unscattered mark tensor
                with shape ``(data_shape[0] + 1, mark_height, mark_width)``
            mark_scattered_shape (list[int]): The scattered mark shape
                ``(data_shape[0] + 1, mark_scattered_height, mark_scattered_width)``

        Returns:
            torch.Tensor: The scattered mark with shape :attr:`mark_scattered_shape`.
        """
        assert mark_scattered_shape[1] >= mark_unscattered.size(1), \
            f'mark_scattered_height={mark_scattered_shape[1]:d}  >=  mark_height={mark_unscattered.size(1):d}'
        assert mark_scattered_shape[2] >= mark_unscattered.size(2), \
            f'mark_scattered_width={mark_scattered_shape[2]:d}  >=  mark_width={mark_unscattered.size(2):d}'
        pixel_num = mark_unscattered[0].numel()
        mark = torch.zeros(mark_scattered_shape, device=env['device'])
        idx = torch.randperm(mark[0].numel())[:pixel_num]
        mark.flatten(1)[:, idx].copy_(mark_unscattered.flatten(1))
        return mark

    # ------------------------------ I/O --------------------------- #

    def load_mark(
        self,
        mark_img: str | Image.Image | np.ndarray | torch.Tensor,
        mark_background_color: None | str | torch.Tensor = 'auto',
        already_processed: bool = False
    ) -> torch.Tensor:
        r"""Load watermark tensor from image :attr:`mark_img`,
        scale by calling :any:`PIL.Image.Image.resize`
        and transform to ``(channel + 1, height, width)`` with alpha channel.

        Args:
            mark_img (PIL.Image.Image | str): Pillow image instance or file path.
            mark_background_color (str | torch.Tensor | None): Mark background color.
                If :class:`str`, choose from ``['auto', 'black', 'white']``;
                else, it shall be 1-dim tensor ranging in ``[0, 1]``.
                It's ignored when alpha channel in watermark image.
                Defaults to ``'auto'``.
            already_processed (bool):
                If ``True``, will just load :attr:`mark_img` as :attr:`self.mark`.
                Defaults to ``False``.

        Returns:
            torch.Tensor: Watermark tensor ranging in ``[0, 1]``
                with shape ``(channel + 1, height, width)`` with alpha channel.
        """
        if isinstance(mark_img, str):
            if mark_img.endswith('.npy'):
                mark_img = np.load(mark_img)
            else:
                if not os.path.isfile(mark_img) and \
                        not os.path.isfile(mark_img := os.path.join(dir_path, mark_img)):
                    raise FileNotFoundError(mark_img.removeprefix(dir_path))
                mark_img = F.convert_image_dtype(F.pil_to_tensor(Image.open(mark_img)))
        if isinstance(mark_img, np.ndarray):
            mark_img = torch.from_numpy(mark_img)
        mark: torch.Tensor = mark_img.to(device=env['device'])
        if not already_processed:
            mark = F.resize(mark, size=(self.mark_width, self.mark_height))
            alpha_mask = torch.ones_like(mark[0])
            if mark.size(0) == 4:
                mark = mark[:-1]
                alpha_mask = mark[-1]
            if self.data_shape[0] == 1 and mark.size(0) == 3:
                mark = F.rgb_to_grayscale(mark, num_output_channels=1)
            mark = torch.cat([mark, alpha_mask.unsqueeze(0)])

            if mark_background_color is not None:
                mark = update_mark_alpha_channel(mark, get_edge_color(mark, mark_background_color))
            if self.mark_random_init:
                mark[:-1] = torch.rand_like(mark[:-1])

            if self.mark_scattered:
                mark_scattered_shape = [mark.size(0), self.mark_scattered_height, self.mark_scattered_width]
                mark = self.scatter_mark(mark, mark_scattered_shape)
        self.mark_height, self.mark_width = mark.shape[-2:]
        self.mark = mark
        return mark


def add_argument(parser: argparse.ArgumentParser) -> argparse._ArgumentGroup:
    r"""
    | Add watermark arguments to argument parser.
    | For specific arguments implementation, see :meth:`Watermark.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
    """
    group = parser.add_argument_group('{yellow}mark{reset}'.format(**ansi))
    return Watermark.add_argument(group)


def create(mark_path: str = None, data_shape: list[int] = None,
           dataset_name: str = None, dataset: str | ImageSet = None,
           config: Config = config, **kwargs) -> Watermark:
    r"""
    | Create a watermark instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | For watermark implementation, see :class:`Watermark`.

    Args:
        mark_path (str):
            | Path to watermark image or npy file.
              There are some preset marks in the package.
            | Defaults to ``'square_white.png'``.
        data_shape (list[int]): The shape of image data ``[C, H, W]``.
        dataset_name (str): The dataset name.
        dataset (str): The alias of `dataset_name`.
        config (Config): The default parameter config.
        **kwargs: Keyword arguments
            passed to dataset init method.

    Returns:
        Watermark: Watermark instance.
    """
    if data_shape is None:
        assert isinstance(dataset, ImageSet), 'Please specify data_shape or dataset'
        data_shape = dataset.data_shape
    if dataset_name is None and dataset is not None:
        dataset_name = dataset.name
    result = config.get_config(dataset_name=dataset_name)[
        'mark'].update(kwargs).update(mark_path=mark_path)
    return Watermark(data_shape=data_shape, **result)



#!/usr/bin/env python3

from trojanzoo.attacks import Attack

from .abstract import *

from .adv import *
from .poison import *
from .backdoor import *

from . import adv, poison, backdoor

from trojanvision.configs import config
import trojanzoo.attacks

import argparse
from trojanvision.datasets import ImageSet
from trojanvision.models import ImageModel
from trojanzoo.configs import Config


module_list = [adv, backdoor, poison]
__all__ = ['add_argument', 'create', 'Attack']
class_dict: dict[str, type[Attack]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)


def add_argument(parser: argparse.ArgumentParser,
                 attack_name: str = None, attack: str | Attack = None,
                 class_dict: dict[str, type[Attack]] = class_dict):
    r"""
    | Add attack arguments to argument parser.
    | For specific arguments implementation, see :func:`trojanzoo.attacks.Attack.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        attack_name (str): The attack name.
        attack (str | Attack): The attack instance or attack name
            (as the alias of `attack_name`).
        class_dict (dict[str, type[Attack]]):
            Map from attack name to attack class.
            Defaults to ``trojanvision.attacks.class_dict``.

    Returns:
        argparse._ArgumentGroup: The argument group.

    See Also:
        :func:`trojanzoo.attacks.add_argument()`
    """
    return trojanzoo.attacks.add_argument(parser=parser, attack_name=attack_name, attack=attack,
                                          class_dict=class_dict)


def create(attack_name: str = None, attack: str | Attack = None,
           dataset_name: str = None, dataset: str | ImageSet = None,
           model_name: str = None, model: str | ImageModel = None,
           config: Config = config, class_dict: dict[str, type[Attack]] = class_dict,
           **kwargs) -> Attack:
    r"""
    | Create an attack instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{attack_dir}/{dataset.data_type}/{dataset.name}/{model.name}/{attack.name}'``.
    | For attack implementation, see :class:`trojanzoo.attacks.Attack`.

    Args:
        attack_name (str): The attack name.
        attack (str | Attack): The attack instance or attack name
            (as the alias of `attack_name`).
        dataset_name (str): The dataset name.
        dataset (str | ImageSet):
            Dataset instance or dataset name
            (as the alias of `dataset_name`).
        model_name (str): The model name.
        model (str | ImageModel): The model instance or model name
            (as the alias of `model_name`).
        config (Config): The default parameter config.
        class_dict (dict[str, type[Attack]]):
            Map from attack name to attack class.
            Defaults to ``trojanvision.attacks.class_dict``.
        **kwargs: Keyword arguments
            passed to attack init method.

    Returns:
        Attack: The attack instance.

    See Also:
        :func:`trojanzoo.attacks.create()`
    """
    return trojanzoo.attacks.create(attack_name=attack_name, attack=attack,
                                    dataset_name=dataset_name, dataset=dataset,
                                    model_name=model_name, model=model,
                                    config=config, class_dict=class_dict,
                                    **kwargs)



#!/usr/bin/env python3

from trojanzoo.models import Model
from trojanzoo.configs import config
from trojanzoo.environ import env
from trojanzoo.utils.model import ExponentialMovingAverage
from trojanzoo.utils.module import get_name, BasicObject
from trojanzoo.utils.output import ansi, prints
from trojanzoo.utils.fim import KFAC, EKFAC

from typing import TYPE_CHECKING
from typing import Any
from trojanzoo.configs import Config    # TODO: python 3.10
from trojanzoo.datasets import Dataset
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
if TYPE_CHECKING:
    pass


class Trainer(BasicObject):
    r"""A dict-like class to contain training arguments
    which supports attribute-like view as well.

    It inherits :class:`trojanzoo.utils.module.BasicObject`.

    Note:
        The most common usage is ``train(**trainer)``.
        See :meth:`keys()` for details.

    Attributes:
        optim_args (dict[str, Any]): optimizer arguments.
        train_args (dict[str, Any]): train function arguments.
        writer_args (dict[str, Any]):
            :any:`torch.utils.tensorboard.writer.SummaryWriter` arguments.
        optimizer (torch.optim.Optimizer): Optimizer instance.
        lr_scheduler (``torch.optim.lr_scheduler._LRScheduler`` | None):
            LR_Scheduler instance.
        model_ema (~trojanzoo.utils.model.ExponentialMovingAverage | None):
            Exponential Moving Average instance.
        pre_conditioner (~trojanzoo.utils.fim.KFAC | ~trojanzoo.utils.fim.EKFAC | None):
            Pre-conditioner instance.
        writer (~torch.utils.tensorboard.writer.SummaryWriter | None):
            Tensorboard summary writer instance.
    """
    name = 'trainer'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        r"""Add trainer arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('--epochs', type=int,
                           help='training epochs '
                           '(default: config[train][epochs])')
        group.add_argument('--resume', type=int,
                           help='resume training from certain epochs '
                           '(default: 0)')
        group.add_argument('--lr', type=float,
                           help='learning rate (default: 0.1)')
        group.add_argument('--parameters',
                           choices=['features', 'classifier', 'full'],
                           help='training parameters '
                           '("features", "classifier", "full") '
                           '(default: "full")')

        group.add_argument('--OptimType',
                           help='optimizer type (default: SGD)')
        group.add_argument('--momentum', type=float,
                           help='momentum passed to Optimizer (default: 0.9)')
        group.add_argument('--weight_decay', type=float,
                           help='weight_decay passed to Optimizer '
                           '(default: 3e-4)')
        group.add_argument('--nesterov', action='store_true',
                           help='enable nesterov for SGD optimizer')

        group.add_argument('--lr_scheduler', action='store_true',
                           help='enable lr scheduler')
        group.add_argument('--lr_scheduler_type',
                           choices=['StepLR', 'CosineAnnealingLR',
                                    'ExponentialLR'],
                           help='the lr scheduler '
                           '(default: CosineAnnealingLR)')
        group.add_argument('--lr_min', type=float,
                           help='min learning rate for `eta_min` '
                           'in CosineAnnealingLR (default: 0.0)')
        group.add_argument('--lr_warmup_epochs', type=int,
                           help='the number of epochs to warmup (default: 0)')
        group.add_argument('--lr_warmup_method',
                           choices=['constant', 'linear'],
                           help='the warmup method (default: constant)')
        group.add_argument('--lr_step_size', type=int,
                           help='decrease lr every step-size epochs '
                           '(default: 30)')
        group.add_argument('--lr_gamma', type=float,
                           help='decrease lr by a factor of lr-gamma '
                           '(default: 0.1)')

        group.add_argument('--model_ema', action='store_true',
                           help='enable tracking Exponential Moving Average '
                           'of model parameters')
        group.add_argument('--model_ema_steps', type=int,
                           help='the number of iterations that controls '
                           'how often to update the EMA model (default: 32)')
        group.add_argument('--model_ema_decay', type=float,
                           help='decay factor for Exponential Moving Average '
                           'of model parameters (default: 0.99998)')

        group.add_argument('--pre_conditioner', choices=[None, 'kfac', 'ekfac'],
                           help='Using kfac/ekfac preconditioner')

        group.add_argument('--amp', action='store_true',
                           help='Use torch.cuda.amp '
                           'for mixed precision training')
        group.add_argument('--grad_clip', type=float,
                           help='Gradient Clipping max norms')

        group.add_argument('--validate_interval', type=int,
                           help='validate interval during training epochs '
                           '(default: 10)')
        group.add_argument('--save', action='store_true',
                           help='save training results')

        group.add_argument('--tensorboard', action='store_true',
                           help='save training logging for tensorboard')
        group.add_argument('--log_dir',
                           help='save training logging for tensorboard')
        group.add_argument('--flush_secs', type=int,
                           help='How often (seconds) '
                           'to flush the pending events '
                           'and summaries to disk')
        return group

    def __init__(self, optim_args: dict[str, Any] = {},
                 train_args: dict[str, Any] = {},
                 writer_args: dict[str, Any] = {},
                 optimizer: Optimizer = None,
                 lr_scheduler: _LRScheduler = None,
                 model_ema: ExponentialMovingAverage = None,
                 pre_conditioner: None | KFAC | EKFAC = None,
                 writer=None, **kwargs):
        super().__init__(**kwargs)
        self.param_list['trainer'] = ['optim_args', 'train_args', 'writer_args',
                                      'optimizer', 'lr_scheduler',
                                      'pre_conditioner', 'model_ema',
                                      'writer']
        # TODO: issue 6 why? to avoid BadAppend issues
        self.optim_args = optim_args.copy()
        self.train_args = train_args.copy()
        self.writer_args = writer_args.copy()
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.model_ema = model_ema
        self.pre_conditioner = pre_conditioner
        self.writer = writer

    def __getitem__(self, key: str):
        if key in self.train_args.keys():
            return self.train_args[key]
        return getattr(self, key)

    def __getattr__(self, key: str):
        if key in self.train_args.keys():
            return self.train_args[key]
        raise AttributeError(key)

    def keys(self) -> list[str]:
        r"""Keys include:

            * | All attributes exclude
              | ``['optim_args', 'train_args', 'writer_args']``
            * train_args

        Returns:
            list[str]: The list of keys.
        """
        keys = self.param_list['trainer'].copy()
        keys.remove('optim_args')
        keys.remove('train_args')
        keys.remove('writer_args')
        keys.extend(self.train_args.keys())
        return keys

    def summary(self, indent: int = None):
        indent = indent if indent is not None else self.indent
        prints('{blue_light}{0:<30s}{reset} Parameters: '.format(
            self.name, **ansi), indent=indent)
        prints('{yellow}{0}{reset}'.format(self.__class__.__name__, **ansi), indent=indent)
        for key in self.param_list['trainer']:
            value = getattr(self, key)
            if value:
                prints('{green}{0:<10s}{reset}'.format(key, **ansi),
                       indent=indent + 10)
                if isinstance(value, dict):
                    value = {k: str(v).split('\n')[0] for k, v in value.items()}
                prints(value, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)


def add_argument(parser: argparse.ArgumentParser,
                 ClassType: type[Trainer] = Trainer
                 ) -> argparse._ArgumentGroup:
    r"""
    | Add trainer arguments to argument parser.
    | For specific arguments implementation, see :meth:`Trainer.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        ClassType (type[Trainer]): The trainer type.
            Defaults to :class:`Trainer`.

    Returns:
        argparse._ArgumentGroup: The argument group.
    """
    group = parser.add_argument_group('{yellow}trainer{reset}'.format(**ansi))
    return ClassType.add_argument(group)


def create(dataset_name: None | str = None,
           dataset: None | str | Dataset = None,
           model: None | Model = None,
           model_ema: None | bool = False,
           pre_conditioner: None | str = None,
           tensorboard: None | bool = None,
           ClassType: type[Trainer] = Trainer,
           config: Config = config, **kwargs):
    r"""
    | Create a trainer instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | For trainer implementation, see :class:`Trainer`.

    Args:
        dataset_name (str): The dataset name.
        dataset (str | trojanzoo.datasets.Dataset):
            Dataset instance
            (required for :attr:`model_ema`)
            or dataset name
            (as the alias of `dataset_name`).
        model (trojanzoo.models.Model): Model instance.
        model_ema (bool): Whether to use
            :class:`~trojanzoo.utils.model.ExponentialMovingAverage`.
            Defaults to ``False``.
        pre_conditioner (str): Choose from

            * ``None``
            * ``'kfac'``: :class:`~trojanzoo.utils.fim.KFAC`
            * ``'ekfac'``: :class:`~trojanzoo.utils.fim.EKFAC`

            Defaults to ``None``.
        tensorboard (bool): Whether to use
            :any:`torch.utils.tensorboard.writer.SummaryWriter`.
            Defaults to ``False``.
        ClassType (type[Trainer]): The trainer class type.
            Defaults to :class:`Trainer`.
        config (Config): The default parameter config.
        **kwargs: The keyword arguments in keys of
            ``['optim_args', 'train_args', 'writer_args']``.

    Returns:
        Trainer: The trainer instance.
    """
    assert isinstance(model, Model)
    dataset_name = get_name(name=dataset_name, module=dataset,
                            arg_list=['-d', '--dataset'])
    result = config.get_config(dataset_name=dataset_name
                               )['trainer'].update(kwargs)

    optim_keys = model.define_optimizer.__code__.co_varnames
    train_keys = model._train.__code__.co_varnames
    optim_args: dict[str, Any] = {}
    train_args: dict[str, Any] = {}
    for key, value in result.items():
        if key in optim_keys:
            optim_args[key] = value
        elif key in train_keys and key != 'verbose':
            train_args[key] = value
    train_args['epochs'] = result['epochs']
    train_args['lr_warmup_epochs'] = result['lr_warmup_epochs']

    optimizer, lr_scheduler = model.define_optimizer(**optim_args)

    module = model._model
    match optim_args['parameters']:
        case 'features':
            module = module.features
        case 'classifier':
            module = module.classifier

    # https://github.com/pytorch/vision/blob/main/references/classification/train.py
    model_ema_module = None
    if model_ema:
        # Decay adjustment that aims to keep the decay independent from other hyper-parameters originally proposed at:
        # https://github.com/facebookresearch/pycls/blob/f8cd9627/pycls/core/net.py#L123
        #
        # total_ema_updates = (Dataset_size / n_GPUs) * epochs / (batch_size_per_gpu * EMA_steps)
        # We consider constant = Dataset_size for a given dataset/setup and ommit it. Thus:
        # adjust = 1 / total_ema_updates ~= n_GPUs * batch_size_per_gpu * EMA_steps / epochs
        adjust = env['world_size'] * dataset.batch_size * \
            result['model_ema_steps'] / result['epochs']
        alpha = 1.0 - result['model_ema_decay']
        alpha = min(1.0, alpha * adjust)
        model_ema_module = ExponentialMovingAverage(
            model._model, decay=1.0 - alpha)

    match pre_conditioner:
        case 'kfac':
            kfac_optimizer = KFAC(module)
        case 'ekfac':
            kfac_optimizer = EKFAC(module)
        case _:
            kfac_optimizer = None

    writer = None
    writer_args: dict[str, Any] = {}
    if tensorboard:
        from torch.utils.tensorboard import SummaryWriter
        # log_dir, flush_secs, ...
        writer_keys = SummaryWriter.__init__.__code__.co_varnames
        for key, value in result.items():
            if key in writer_keys:
                writer_args[key] = value
        writer = SummaryWriter(**writer_args)
    return ClassType(optim_args=optim_args, train_args=train_args,
                     writer_args=writer_args,
                     optimizer=optimizer, lr_scheduler=lr_scheduler,
                     model_ema=model_ema_module,
                     pre_conditioner=kfac_optimizer, writer=writer)



#!/usr/bin/env python3

import trojanvision
from trojanvision.attacks import BadNet
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    trojanvision.environ.add_argument(parser)
    trojanvision.datasets.add_argument(parser)
    trojanvision.models.add_argument(parser)
    trojanvision.marks.add_argument(parser)
    trojanvision.attacks.add_argument(parser)
    kwargs = parser.parse_args().__dict__

    env = trojanvision.environ.create(**kwargs)
    dataset = trojanvision.datasets.create(**kwargs)
    model = trojanvision.models.create(dataset=dataset, **kwargs)
    mark = trojanvision.marks.create(dataset=dataset, **kwargs)
    attack: BadNet = trojanvision.attacks.create(dataset=dataset, model=model, mark=mark, **kwargs)

    if env['verbose']:
        trojanvision.summary(env=env, dataset=dataset, model=model, mark=mark, attack=attack)
    attack.load()
    attack.validate_fn()



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --tqdm --pretrained --validate_interval 1 --epochs 10 --lr 0.01 --mark_random_init --attack trojannn

CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --tqdm --pretrained --validate_interval 1 --epochs 10 --lr 0.01 --mark_random_init --attack trojannn --model vgg13_comp --preprocess_layer classifier.fc1 --preprocess_next_layer classifier.fc2
"""  # noqa: E501

from ...abstract import BackdoorAttack

from trojanvision.environ import env
from trojanzoo.utils.tensor import tanh_func

import torch
import torch.nn.functional as F
import numpy as np
import skimage.restoration

import argparse


class TrojanNN(BackdoorAttack):
    r"""TrojanNN proposed by Yingqi Liu from Purdue University in NDSS 2018.

    It inherits :class:`trojanvision.attacks.BackdoorAttack`.

    Based on :class:`trojanvision.attacks.BadNet`,
    TrojanNN preprocesses watermark pixel values to maximize
    activations of well-connected neurons in :attr:`self.preprocess_layer`.

    See Also:
        * paper: `Trojaning Attack on Neural Networks`_
        * code: https://github.com/PurduePAML/TrojanNN
        * website: https://purduepaml.github.io/TrojanNN

    Args:
        preprocess_layer (str): The chosen layer
            to maximize neuron activation.
            Defaults to ``'flatten'``.
        preprocess_next_layer (str): The next layer
            after preprocess_layer to find neuron index.
            Defaults to ``'classifier.fc'``.
        target_value (float): TrojanNN neuron activation target value.
            Defaults to ``100.0``.
        neuron_num (int): TrojanNN neuron number to maximize activation.
            Defaults to ``2``.
        neuron_epoch (int): TrojanNN neuron optimization epoch.
            Defaults to ``1000``.
        neuron_lr (float): TrojanNN neuron optimization learning rate.
            Defaults to ``0.1``.

    .. _Trojaning Attack on Neural Networks:
        https://github.com/PurduePAML/TrojanNN/blob/master/trojan_nn.pdf
    """

    name: str = 'trojannn'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--preprocess_layer',
                           help='the chosen layer to maximize neuron activation '
                           '(default: "flatten")')
        group.add_argument('--preprocess_next_layer',
                           help='the next layer after preprocess_layer to find neuron index '
                           '(default: "classifier.fc")')
        group.add_argument('--target_value', type=float,
                           help='trojannn neuron activation target value '
                           '(default: 100)')
        group.add_argument('--neuron_num', type=int,
                           help='Trojan Net neuron numbers in neuron preprocessing '
                           '(default: 2)')
        group.add_argument('--neuron_epoch', type=int,
                           help='trojann neuron optimization epoch '
                           '(default: 1000)')
        group.add_argument('--neuron_lr', type=float,
                           help='trojann neuron optimization learning rate '
                           '(default: 0.1)')
        return group

    def __init__(self, preprocess_layer: str = 'flatten', preprocess_next_layer: str = 'classifier.fc',
                 target_value: float = 100.0, neuron_num: int = 2,
                 neuron_lr: float = 0.1, neuron_epoch: int = 1000,
                 **kwargs):
        super().__init__(**kwargs)
        if self.mark.mark_random_pos:
            raise Exception('TrojanNN requires "mark_random_pos" to be False to max activate neurons.')

        self.param_list['trojannn'] = ['preprocess_layer', 'preprocess_next_layer',
                                       'target_value', 'neuron_num',
                                       'neuron_lr', 'neuron_epoch']
        self.preprocess_layer = preprocess_layer
        self.preprocess_next_layer = preprocess_next_layer
        self.target_value = target_value

        self.neuron_lr = neuron_lr
        self.neuron_epoch = neuron_epoch
        self.neuron_num = neuron_num

        self.neuron_idx: torch.Tensor = None
        self.background = torch.zeros(self.dataset.data_shape, device=env['device']).unsqueeze(0)
        # Original code: doesn't work on resnet18_comp
        # self.background = torch.normal(mean=175.0 / 255, std=8.0 / 255,
        #                                size=self.dataset.data_shape,
        #                                device=env['device']).clamp(0, 1).unsqueeze(0)

    def attack(self, *args, **kwargs):
        self.neuron_idx = self.get_neuron_idx()
        print('Neuron Index: ', self.neuron_idx.cpu().tolist())
        self.preprocess_mark(neuron_idx=self.neuron_idx)
        return super().attack(*args, **kwargs)

    def get_neuron_idx(self) -> torch.Tensor:
        r"""Get top :attr:`self.neuron_num` well-connected neurons
        in :attr:`self.preprocess_layer`.

        It is calculated w.r.t. in_channels of
        :attr:`self.preprocess_next_layer` weights.

        Returns:
            torch.Tensor: Neuron index list tensor with shape ``(self.neuron_num)``.
        """
        weight = self.model.state_dict()[self.preprocess_next_layer + '.weight'].abs()
        if weight.dim() > 2:
            weight = weight.flatten(2).sum(2)
        return weight.sum(0).argsort(descending=True)[:self.neuron_num]

    def get_neuron_value(self, trigger_input: torch.Tensor, neuron_idx: torch.Tensor) -> float:
        r"""Get average neuron activation value of :attr:`trigger_input` for :attr:`neuron_idx`.

        The feature map is obtained by calling :meth:`trojanzoo.models.Model.get_layer()`.

        Args:
            trigger_input (torch.Tensor): Poison input tensor with shape ``(N, C, H, W)``.
            neuron_idx (torch.Tensor): Neuron index list tensor with shape ``(self.neuron_num)``.

        Returns:
            float: Average neuron activation value.
        """
        trigger_feats = self.model.get_layer(
            trigger_input, layer_output=self.preprocess_layer)[:, neuron_idx].abs()
        if trigger_feats.dim() > 2:
            trigger_feats = trigger_feats.flatten(2).sum(2)
        return trigger_feats.sum().item()

    def preprocess_mark(self, neuron_idx: torch.Tensor):
        r"""Optimize mark to maxmize activation on :attr:`neuron_idx`.
        It uses :any:`torch.optim.Adam` and
        :any:`torch.optim.lr_scheduler.CosineAnnealingLR`
        with tanh objective funcion.

        The feature map is obtained by calling
        :meth:`trojanvision.models.ImageModel.get_layer()`.

        Args:
            neuron_idx (torch.Tensor): Neuron index list tensor with shape ``(self.neuron_num)``.
        """
        atanh_mark = torch.randn_like(self.mark.mark[:-1], requires_grad=True)
        # Original code: no difference
        # start_h, start_w = self.mark.mark_height_offset, self.mark.mark_width_offset
        # end_h, end_w = start_h + self.mark.mark_height, start_w + self.mark.mark_width
        # self.mark.mark[:-1] = self.background[0, :, start_h:end_h, start_w:end_w]
        # atanh_mark = (self.mark.mark[:-1] * (2 - 1e-5) - 1).atanh()
        # atanh_mark.requires_grad_()
        self.mark.mark[:-1] = tanh_func(atanh_mark.detach())
        self.mark.mark.detach_()

        optimizer = torch.optim.Adam([atanh_mark], lr=self.neuron_lr)
        # No difference for SGD
        # optimizer = optim.SGD([atanh_mark], lr=self.neuron_lr)
        optimizer.zero_grad()
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=self.neuron_epoch)

        with torch.no_grad():
            trigger_input = self.add_mark(self.background, mark_alpha=1.0)
            print('Neuron Value Before Preprocessing:',
                  f'{self.get_neuron_value(trigger_input, neuron_idx):.5f}')

        for _ in range(self.neuron_epoch):
            self.mark.mark[:-1] = tanh_func(atanh_mark)
            trigger_input = self.add_mark(self.background, mark_alpha=1.0)
            trigger_feats = self.model.get_layer(trigger_input, layer_output=self.preprocess_layer)
            trigger_feats = trigger_feats[:, neuron_idx].abs()
            if trigger_feats.dim() > 2:
                trigger_feats = trigger_feats.flatten(2).sum(2)
                # Original code
                # trigger_feats = trigger_feats.flatten(2).amax(2)
            loss = F.mse_loss(trigger_feats, self.target_value * torch.ones_like(trigger_feats),
                              reduction='sum')   # paper's formula
            # Original code: no difference
            # loss = -self.target_value * trigger_feats.sum()
            loss.backward(inputs=[atanh_mark])
            optimizer.step()
            lr_scheduler.step()
            optimizer.zero_grad()
            self.mark.mark.detach_()

            # Original Code: no difference
            # self.mark.mark[:-1] = tanh_func(atanh_mark.detach())
            # trigger = self.denoise(self.add_mark(torch.zeros_like(self.background), mark_alpha=1.0)[0])
            # mark = trigger[:, start_h:end_h, start_w:end_w].clamp(0, 1)
            # atanh_mark.data = (mark * (2 - 1e-5) - 1).atanh()

        atanh_mark.requires_grad_(False)
        self.mark.mark[:-1] = tanh_func(atanh_mark)
        self.mark.mark.detach_()

    # def validate_fn(self, **kwargs) -> tuple[float, float]:
    #     if self.neuron_idx is not None:
    #         with torch.no_grad():
    #             trigger_input = self.add_mark(self.background, mark_alpha=1.0)
    #             print(f'Neuron Value: {self.get_neuron_value(trigger_input, self.neuron_idx):.5f}')
    #     return super().validate_fn(**kwargs)

    @staticmethod
    def denoise(img: torch.Tensor, weight: float = 1.0,
                max_num_iter: int = 100, eps: float = 1e-3) -> torch.Tensor:
        r"""Denoise image by calling :any:`skimage.restoration.denoise_tv_bregman`.

        Warning:
            This method is currently unused in :meth:`preprocess_mark()`
            because no performance difference is observed.

        Args:
            img (torch.Tensor): Noisy image tensor with shape ``(C, H, W)``.

        Returns:
            torch.Tensor: Denoised image tensor with shape ``(C, H, W)``.
        """
        if img.size(0) == 1:
            img_np: np.ndarray = img[0].detach().cpu().numpy()
        else:
            img_np = img.detach().cpu().permute(1, 2, 0).contiguous().numpy()

        denoised_img_np = skimage.restoration.denoise_tv_bregman(
            img_np, weight=weight, max_num_iter=max_num_iter, eps=eps)
        denoised_img = torch.from_numpy(denoised_img_np)

        if denoised_img.dim() == 2:
            denoised_img.unsqueeze_(0)
        else:
            denoised_img = denoised_img.permute(2, 0, 1).contiguous()
        return img.to(device=img.device)



#!/usr/bin/env python3

from ...abstract import DynamicBackdoor

from .input_aware_dynamic import InputAwareDynamic


__all__ = ['InputAwareDynamic']

class_dict: dict[str, type[DynamicBackdoor]] = {
    'input_aware_dynamic': InputAwareDynamic,
}



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --epochs 10 --lr 0.01 --mark_random_init --attack imc
"""  # noqa: E501

from .trojannn import TrojanNN
from trojanzoo.utils.tensor import tanh_func

import torch
import torch.optim as optim
import argparse

from collections.abc import Callable


class IMC(TrojanNN):
    r"""Input Model Co-optimization (IMC) proposed by Ren Pang
    from Pennsylvania State University in CCS 2020.

    It inherits :class:`trojanvision.attacks.BackdoorAttack`.

    Based on :class:`trojanvision.attacks.TrojanNN`,
    IMC optimizes the watermark using Adam optimizer during model retraining.

    See Also:
        * paper: `A Tale of Evil Twins\: Adversarial Inputs versus Poisoned Models`_
        * code: TrojanZoo is the official implementation of IMC ^_^

    Args:
        attack_remask_epochs (int): Inner epoch to optimize watermark during each training epoch.
            Defaults to ``20``.
        attack_remask_lr (float): Learning rate of Adam optimizer to optimize watermark.
            Defaults to ``0.1``.

    .. _A Tale of Evil Twins\: Adversarial Inputs versus Poisoned Models:
        https://arxiv.org/abs/1911.01559
    """  # noqa: E501

    name: str = 'imc'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--attack_remask_epochs', type=int,
                           help='inner epoch to optimize watermark during each training epoch '
                           '(default: 1)')
        group.add_argument('--attack_remask_lr', type=float,
                           help='learning rate of Adam optimizer to optimize watermark'
                           '(default: 0.1)')
        return group

    def __init__(self, attack_remask_epochs: int = 1, attack_remask_lr: float = 0.1, **kwargs):
        super().__init__(**kwargs)
        self.param_list['imc'] = ['attack_remask_epochs', 'attack_remask_lr']
        self.attack_remask_epochs = attack_remask_epochs
        self.attack_remask_lr = attack_remask_lr

    def attack(self, epochs: int, **kwargs):
        return super().attack(epochs, epoch_fn=self.epoch_fn, **kwargs)

    def epoch_fn(self, **kwargs):
        self.optimize_mark()

    def optimize_mark(self, loss_fn: Callable[..., torch.Tensor] = None, **kwargs):
        r"""Optimize watermark at the beginning of each training epoch."""
        loss_fn = loss_fn or self.model.loss

        atanh_mark = torch.randn_like(self.mark.mark[:-1], requires_grad=True)
        optimizer = optim.Adam([atanh_mark], lr=self.attack_remask_lr)
        optimizer.zero_grad()

        for _ in range(self.attack_remask_epochs):
            for data in self.dataset.loader['train']:
                self.mark.mark[:-1] = tanh_func(atanh_mark)
                _input, _label = self.model.get_data(data)
                trigger_input = self.add_mark(_input)
                trigger_label = self.target_class * torch.ones_like(_label)
                loss = loss_fn(trigger_input, trigger_label)
                loss.backward(inputs=[atanh_mark])
                optimizer.step()
                optimizer.zero_grad()
                self.mark.mark.detach_()
        atanh_mark.requires_grad_(False)
        self.mark.mark[:-1] = tanh_func(atanh_mark)
        self.mark.mark.detach_()



#!/usr/bin/env python3

from ...abstract import ModelInspection
from trojanvision.environ import env
from trojanzoo.utils.data import sample_batch
from trojanzoo.utils.metric import mask_jaccard
from trojanzoo.utils.output import ansi, output_iter, prints

import torch
import numpy as np
import argparse
import os

import torch.utils.data


# use_mask: bool = True
# count_mask: bool = True
# ImageNet arguments from https://github.com/naiyeleo/ABS/blob/master/TrojAI_competition/round1/abs_pytorch_round1.py
# {
#     'samp_k': 2,
#     'defense_remask_lr': 0.4,
#     'defense_remask_epoch': 50,
#     'remask_weight': 50,
#     'max_troj_size': 400,
#     'top_n_neurons': 10,
# }


class ABS(ModelInspection):
    r"""Artificial Brain Stimulation proposed by Yingqi Liu
    from Purdue University in CCS 2019.

    It is a model inspection backdoor defense
    that inherits :class:`trojanvision.defenses.ModelInspection`.

    See Also:
        * paper: `ABS\: Scanning Neural Networks for Back-doors by Artificial Brain Stimulation`_
        * code: https://github.com/naiyeleo/ABS

    .. _ABS\: Scanning Neural Networks for Back-doors by Artificial Brain Stimulation:
        https://openreview.net/forum?id=YHWF1F1RBgF
    """
    name: str = 'abs'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--seed_data_num', type=int,
                           help='abs seed data number (default: -5)')
        group.add_argument('--mask_eps', type=float,
                           help='abs mask epsilon threshold (default: 0.01)')
        group.add_argument('--samp_k', type=int,
                           help='abs sample k multiplier in neuron sampling (default: 8)')
        group.add_argument('--same_range', action='store_true',
                           help='abs same range in neuron sampling (default: False)')
        group.add_argument('--n_samples', type=int,
                           help='abs n_samples in neuron sampling (default: 5)')
        group.add_argument('--top_n_neurons', type=int,
                           help='abs top-n neuron number in neuron sampling (default: 20)')
        group.add_argument('--max_troj_size', type=int,
                           help='abs max trojan trigger size in pixel number (default: 16)')
        group.add_argument('--remask_weight', type=float,
                           help='abs optimization norm loss weight (default: 500)')
        return group

    def __init__(self, seed_data_num: int = -5, mask_eps: float = 0.01,
                 samp_k: int = 8, same_range: bool = False,
                 n_samples: int = 5, top_n_neurons: int = 20,
                 max_troj_size: int = 16, remask_weight: float = 500.0,
                 defense_remask_lr: float = 0.1,
                 defense_remask_epoch: int = 1000,
                 **kwargs):
        super().__init__(defense_remask_epoch=defense_remask_epoch,
                         defense_remask_lr=defense_remask_lr,
                         cost=0.0, **kwargs)
        self.param_list['abs'] = ['seed_data_num', 'mask_eps',
                                  'samp_k', 'same_range',
                                  'n_samples', 'top_n_neurons',
                                  'max_troj_size', 'remask_weight']

        if seed_data_num < 0:
            seed_data_num = self.model.num_classes * abs(seed_data_num)
        self.seed_data_num = seed_data_num

        self.mask_eps = mask_eps
        # -----------Neural Sampling------------- #
        self.samp_k = samp_k
        self.same_range = same_range
        self.n_samples = n_samples
        self.top_n_neurons = top_n_neurons

        self.max_troj_size = max_troj_size
        self.remask_weight = remask_weight

        self.seed_data = self.get_seed_data()
        self.loader = [(self.seed_data['input'], self.seed_data['label'])]

    def get_mark_loss_list(self) -> tuple[torch.Tensor, list[float], list[float]]:
        print('sample neurons')
        all_ps = self.sample_neuron(self.seed_data['input'])
        print('find min max')
        self.neuron_dict = self.find_min_max(all_ps, self.seed_data['label'])

        format_str = self.serialize_format(layer='20s', neuron='5d', value='10.3f')
        # Output neuron dict information
        for label in range(self.model.num_classes):
            print('Class: ', output_iter(label, self.model.num_classes))
            for _dict in reversed(self.neuron_dict[label]):
                prints(format_str.format(**_dict), indent=4)
        print()
        print('optimize marks')
        return super().get_mark_loss_list(verbose=False)

    def optimize_mark(self, label: int, **kwargs) -> tuple[torch.Tensor, float]:
        format_dict = dict(layer='20s', neuron='5d', value='10.3f',
                           loss='10.3f', asr='8.3f', norm='8.3f')
        if not self.attack.mark.mark_random_pos:
            format_dict['jaccard'] = '5.3f'
            select_num = self.attack.mark.mark_height * self.attack.mark.mark_width
        format_str = self.serialize_format(**format_dict)
        mark_best: torch.Tensor = torch.ones_like(self.attack.mark.mark)
        loss_best: float = 1e7
        asr_best: float = 0.0
        dict_best = {}
        for _dict in reversed(self.neuron_dict[label]):
            mark, loss = super().optimize_mark(label, loader=self.loader,
                                               verbose=False, **_dict)
            _dict['mark'] = mark.detach().cpu().clone().numpy()
            asr, _ = self.model._validate(get_data_fn=self.attack.get_data,
                                          keep_org=False, verbose=False)
            norm = float(mark[-1].flatten().norm(p=1))
            str_dict = dict(loss=loss, asr=asr, norm=norm, **_dict)
            if not self.attack.mark.mark_random_pos:
                overlap = mask_jaccard(self.attack.mark.get_mask(),
                                       self.real_mask,
                                       select_num=select_num)
                str_dict['jaccard'] = overlap
            prints(format_str.format(**str_dict), indent=4)
            if asr > asr_best:
                asr_best = asr
                mark_best = mark
                loss_best = loss
                dict_best = str_dict
        format_str = self.serialize_format(color='yellow', **format_dict)
        print()
        prints(format_str.format(**dict_best), indent=4)
        self.attack.mark.mark = mark_best
        return mark_best, loss_best

    def loss(self, _input: torch.Tensor, _label: torch.Tensor,
             layer: str, neuron: int, **kwargs) -> torch.Tensor:
        trigger_input = self.attack.add_mark(_input)
        feats = self.model.get_layer(trigger_input, layer_output=layer)
        feats = feats + (feats > 0) * feats    # sum up with relu
        vloss1 = feats[:, neuron].sum()
        vloss2 = feats.sum() - vloss1

        # if not use_mask:
        #     tvloss = total_variation(self.attack.mark.mark[:-1])
        #     ssim_loss = - self.ssim(X, _input)
        #     ssim_loss *= 10 if ssim_loss < -2 else 10000
        #     loss = -vloss1 + 1e-5 * vloss2 + 1e-3 * tvloss
        #     loss = 0.01 * loss + ssim_loss

        norm_loss = self.attack.mark.mark[-1].sum()
        mask_nz = float((self.attack.mark.mark[-1] > self.mask_eps).sum())
        if mask_nz > 1.2 * self.max_troj_size:
            norm_loss *= 2 * self.remask_weight
        elif mask_nz > self.max_troj_size:
            norm_loss *= self.remask_weight
        else:
            norm_loss *= 0.01
        return -vloss1 + 1e-4 * vloss2 + norm_loss

    # ---------------------------- Seed Data --------------------------- #
    def gen_seed_data(self) -> dict[str, np.ndarray]:
        r"""Generate seed data.

        Returns:
            dict[str, numpy.ndarray]:
                Seed data dict with keys ``'input'`` and ``'label'``.
        """
        torch.manual_seed(env['seed'])
        if self.seed_data_num % self.model.num_classes:
            raise ValueError(
                f'seed_data_num({self.seed_data_num:d}) % num_classes({self.model.num_classes:d}) should be 0.')
        seed_class_num: int = self.seed_data_num // self.model.num_classes
        x, y = [], []
        for _class in range(self.model.num_classes):
            class_set = self.dataset.get_dataset(mode='train', class_list=[_class])
            _input, _label = sample_batch(class_set, batch_size=seed_class_num)
            x.append(_input)
            y.append(_label)
        x = torch.cat(x).numpy()
        y = torch.cat(y).numpy()
        seed_data = {'input': x, 'label': y}
        seed_path = os.path.join(self.folder_path, f'seed_{self.seed_data_num}.npz')
        np.savez(seed_path, **seed_data)
        print('seed data saved at: ', seed_path)
        return seed_data

    def get_seed_data(self) -> dict[str, torch.Tensor]:
        r"""Get seed data. If npz file doesn't exist,
        call :meth:`gen_seed_data()` to generate.
        """
        seed_path = os.path.join(self.folder_path, f'seed_{self.seed_data_num}.npz')
        seed_data: dict[str, torch.Tensor] = {}
        seed_data_np = dict(np.load(seed_path)) if os.path.exists(seed_path) \
            else self.gen_seed_data()
        seed_data['input'] = torch.from_numpy(seed_data_np['input']).to(device=env['device'])
        seed_data['label'] = torch.from_numpy(seed_data_np['label']).to(device=env['device'], dtype=torch.long)
        return seed_data

    # -----------------------Neural Sample---------------------------- #

    def sample_neuron(self, _input: torch.Tensor) -> dict[str, torch.Tensor]:
        all_ps: dict[str, torch.Tensor] = {}
        layer_output = self.model.get_all_layer(_input, depth=2)
        for layer in layer_output.keys():
            if not layer.startswith('features.') and not layer.startswith('classifier.'):
                continue
            cur_layer_output: torch.Tensor = layer_output[layer].detach().cpu()  # (batch_size, C', H', W')
            channel_num: int = cur_layer_output.shape[1]  # channels

            h_t: torch.Tensor = cur_layer_output.expand([channel_num, self.n_samples] + [-1] * cur_layer_output.dim()
                                                        ).clone()
            # (C, n_samples, batch_size, C', H', W')

            vs = self.samp_k * torch.arange(self.n_samples, device=h_t.device, dtype=torch.float)
            if not self.same_range:
                maxes = cur_layer_output.max()
                vs *= float(maxes) / self.n_samples
            vs_shape = [1] * cur_layer_output.dim()
            vs_shape[0] = -1
            vs = vs.view(vs_shape)  # (n_samples, 1, 1, 1)
            # todo: might use parallel to avoid for loop (torch.Tensor.scatter?)
            for neuron in range(channel_num):
                h_t[neuron, :, :, neuron] = vs
            # todo: the shape is too large
            # result = self.model.get_layer(h_t.flatten(end_dim=2), layer_input=layer).detach().cpu()
            result = []
            for h in h_t:
                h: torch.Tensor
                h = h.to(device=env['device'])
                result.append(self.model.get_layer(h.flatten(end_dim=1), layer_input=layer
                                                   ).detach().cpu())
            result = torch.cat(result)

            result_shape = list(h_t.shape)[:3]
            result_shape.extend(list(result.shape)[1:])
            result = result.view(result_shape)
            all_ps[layer] = result  # (C, n_samples, batch_size, num_classes)
        return all_ps

    def find_min_max(self, all_ps: dict[str, torch.Tensor], _label: torch.Tensor
                     ) -> dict[int, list[dict[str, str | int | float]]]:
        neuron_dict: dict[int, list[dict[str, str | int | float]]]
        neuron_dict = {i: [] for i in range(self.model.num_classes)}
        _label = _label.cpu()
        for layer in all_ps.keys():
            ps = all_ps[layer]  # (C, n_samples, batch_size, num_classes)
            vs: torch.Tensor = ps[:, self.n_samples // 5:].amax(dim=1) \
                - ps[:, :self.n_samples // 5].amin(dim=1)  # (C, batch_size, num_classes)
            values, labels = vs.sort(dim=-1, descending=True)
            condition1 = labels[..., 0].eq(_label)  # exclude the ground-truth labels
            values = torch.where(condition1, values[..., 1] - values[..., 2],
                                 values[..., 0] - values[..., 1])  # (C, batch_size)
            labels = torch.where(condition1, labels[..., 1], labels[..., 0])  # (C, batch_size)

            mode_labels = labels.mode(keepdim=True)[0]  # (C, 1) The most frequent label
            mode_idx = labels.eq(mode_labels)  # (C, batch_size)
            mode_labels_counts = mode_idx.sum(dim=-1)  # (C)
            condition2 = mode_labels_counts.ge(self.seed_data_num * 0.75)
            idx_list = condition2.nonzero().flatten().tolist()
            idx_list = sorted(idx_list, key=lambda idx: float(values[idx][mode_idx[idx]].min()))[:self.top_n_neurons]
            for idx in idx_list:
                value = float(values[idx][mode_idx[idx]].min())
                neuron_dict[int(mode_labels[idx])].append(
                    {'layer': layer, 'neuron': int(idx), 'value': value})
        return neuron_dict

    @staticmethod
    def serialize_format(color: str = 'green', **kwargs: str) -> str:
        _str = ''
        for k, v in kwargs.items():
            _str += '{color}{k}{reset}: {{{k}:{v}}}    '.format(k=k, v=v, color=ansi[color], reset=ansi['reset'])
        return _str.removesuffix('    ')

    # ---------------------------------- Unused ------------------------------- #
    # def filter_img(self):
    #     h, w = self.dataset.data_shape[1:]
    #     mask = torch.zeros(h, w)
    #     mask[2:7, 2:7] = 1
    #     return mask.to(device=env['device'])

    # def nc_filter_img(self) -> torch.Tensor:
    #     h, w = self.dataset.data_shape[1:]
    #     mask = torch.ones(h, w)
    #     return mask.to(device=env['device'])
    #     # TODO: fix
    #     # mask = torch.zeros(h, w)
    #     # if self.use_mask:
    #     #     mask[math.ceil(0.25 * w): math.floor(0.75 * w), math.ceil(0.25 * h): math.floor(0.75 * h)] = 1
    #     # else:
    #     #     mask.add_(1)



#!/usr/bin/env python3

from trojanzoo.configs import config
from trojanzoo.environ import env
from trojanzoo.utils.data import get_class_subset, split_dataset
from trojanzoo.utils.module import BasicObject, get_name
from trojanzoo.utils.output import ansi

import torch
import numpy as np
import functools
import os
from abc import ABC, abstractmethod

from typing import TYPE_CHECKING
from typing import Iterable
from trojanzoo.configs import Config
import argparse    # TODO: python 3.10
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data


class Dataset(ABC, BasicObject):
    r"""
    | An abstract class representing a dataset.
    | It inherits :class:`trojanzoo.utils.module.BasicObject`.

    Note:
        This is the implementation of dataset.
        For users, please use :func:`create` instead, which is more user-friendly.

    Args:
        batch_size (int): Batch size of training set
            (negative number means batch size for each gpu).
        valid_batch_size (int): Batch size of validation set.
            Defaults to ``100``.
        folder_path (str): Folder path to store dataset.
            Defaults to ``None``.

            Note:
                :attr:`folder_path` is usually
                ``'{data_dir}/{data_type}/{name}'``,
                which is claimed as the default value of :func:`create()`.
        download (bool): Download dataset if not exist. Defaults to ``False``.
        split_ratio (float):
            | Split training set for training and validation
              if :attr:`valid_set` is ``False``.
            | The ratio stands for
              :math:`\frac{\text{\# training\ subset}}{\text{\# total\ training\ set}}`.
            | Defaults to ``0.8``.
        num_workers (int): Used in :meth:`get_dataloader()`.
            Defaults to ``4``.
        loss_weights (bool | np.ndarray | torch.Tensor):
            | The loss weights w.r.t. each class.
            | if :any:`numpy.ndarray` or :any:`torch.Tensor`,
              directly set as :attr:`loss_weights` (cpu tensor).
            | if ``True``, set :attr:`loss_weights` as :meth:`get_loss_weights()`;
            | if ``False``, set :attr:`loss_weights` as ``None``.
        **kwargs: Any keyword argument (unused).

    Attributes:
        name (str): Dataset Name. (need overriding)
        loader(dict[str, ~torch.utils.data.DataLoader]):
            | Preset dataloader for users at dataset initialization.
            | It contains ``'train'`` and ``'valid'`` loaders.
        batch_size (int): Batch size of training set (always positive).
        valid_batch_size (int): Batch size of validation set.
            Defaults to ``100``.
        num_classes (int): Number of classes. (need overriding)
        folder_path (str): Folder path to store dataset.
            Defaults to ``None``.

        data_type (str): Data type (e.g., ``'image'``). (need overriding)
        label_names (list[int]): Number of classes. (optional)
        valid_set (bool): Whether having a native validation set.
            Defaults to ``True``.
        split_ratio (float):
            | Split training set for training and validation
              if :attr:`valid_set` is ``False``.
            | The ratio stands for
              :math:`\frac{\text{\# training\ subset}}{\text{\# total\ training\ set}}`.
            | Defaults to ``0.8``.
        loss_weights (torch.Tensor | None): The loss weights w.r.t. each class.
        num_workers (int): Used in :meth:`get_dataloader()`.
            Defaults to ``4``.
        collate_fn (~collections.abc.Callable | None):
            Used in :meth:`get_dataloader()`.
            Defaults to ``None``.
    """
    name = 'dataset'
    data_type: str = None
    num_classes: int = None
    label_names: list[int] = None
    valid_set = True

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        r"""Add dataset arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            The concrete dataset class may override this method to add more arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('-d', '--dataset', dest='dataset_name',
                           help='dataset name (lowercase)')
        group.add_argument('--batch_size', type=int,
                           help='batch size (negative number means '
                           'batch_size for each gpu)')
        group.add_argument('--valid_batch_size', type=int,
                           help='valid batch size')
        group.add_argument('--num_workers', type=int,
                           help='num_workers passed to '
                           'torch.utils.data.DataLoader (default: 4)')
        group.add_argument('--download', action='store_true',
                           help='download dataset if not exist by calling '
                           'self.initialize()')
        group.add_argument('--data_dir', help='directory to contain datasets')
        return group

    def __init__(self, batch_size: int = None,
                 valid_batch_size: int = 100,
                 folder_path: str = None, download: bool = False,
                 split_ratio: float = 0.8, num_workers: int = 4,
                 loss_weights: bool | np.ndarray | torch.Tensor = False,
                 **kwargs):
        super().__init__(**kwargs)
        self.param_list['dataset'] = ['num_classes', 'batch_size', 'valid_batch_size',
                                      'folder_path', 'num_workers', ]
        if not self.valid_set:
            self.param_list['dataset'].append('split_ratio')
        self.__batch_size = batch_size
        self.valid_batch_size = valid_batch_size
        self.split_ratio = split_ratio
        self.num_workers = num_workers
        self.collate_fn: Callable[[Iterable[torch.Tensor]], Iterable[torch.Tensor]] = None
        # ----------------------------------------------- #

        if folder_path is not None:
            folder_path = os.path.normpath(folder_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
        self.folder_path = folder_path
        # ----------------------------------------------- #
        if download and not self.check_files():
            self.initialize()
        # Preset Loader
        try:
            self.loader: dict[str, torch.utils.data.DataLoader] = {}
            self.loader['train'] = self.get_dataloader(mode='train')
            self.loader['valid'] = self.get_dataloader(mode='valid')
        except Exception:
            print(f'Dataset Folder Path: {self.folder_path}')
            raise
        # ----------------------------------------------- #
        # Loss Weights
        match loss_weights:
            case bool():
                # TODO: issue 5 pylance
                loss_weights = self.get_loss_weights() if loss_weights else None
            case np.ndarray():
                loss_weights = torch.from_numpy(loss_weights).to(device=env['device'])
            case torch.Tensor():
                loss_weights = loss_weights.to(device=env['device'])
            case _:
                raise TypeError(type(loss_weights))
        self.loss_weights: None | torch.Tensor = loss_weights

    @functools.cached_property
    def batch_size(self):
        return self.__batch_size if self.__batch_size >= 0 else \
            -self.__batch_size * max(1, env['num_gpus'])

    # TODO: should it be abstractmethod?
    def initialize(self, *args, **kwargs):
        r"""Initialize the dataset (download and extract) if it's not prepared yet
        (need overriding).
        """
        raise NotImplementedError()

    def check_files(self, **kwargs) -> bool:
        r"""Check if the dataset files are prepared.

        Args:
            **kwargs: Keyword arguments passed to :meth:`get_org_dataset`.

        Returns:
            bool: Whether the dataset files are prepared.
        """
        try:
            self.get_org_dataset(mode='train', transform=None, **kwargs)
            if self.valid_set:
                self.get_org_dataset(
                    mode='valid', transform=None, **kwargs)
            return True
        except Exception:
            return False

    @abstractmethod
    def get_transform(self, mode: str) -> Callable:
        r"""Get dataset transform for mode.

        Args:
            mode (str): Dataset mode (e.g., ``'train'`` or ``'valid'``).

        Returns:
            ~collections.abc.Callable: A callable transform.
        """
        ...

    def get_data(self, data, **kwargs):
        r"""Process data. Defaults to directly return :attr:`data`.

        Args:
            data (Any): Unprocessed data.
            **kwargs: Keyword arguments to process data.

        Returns:
            Any: Processed data.
        """
        return data

    def get_org_dataset(self, mode: str,
                        **kwargs) -> torch.utils.data.Dataset:
        r"""Get original dataset that is not splitted.

        Note:
            This is a wrapper and the specific implementation
            is in :meth:`_get_org_dataset`, which needs overriding.

        Args:
            mode (str): Dataset mode (e.g., ``'train'`` or ``'valid'``).
            transform (~collections.abc.Callable):
                The transform applied on dataset.
                Defaults to :meth:`get_transform()`.
            **kwargs: Keyword arguments passed to :meth:`_get_org_dataset`.

        Returns:
            torch.utils.data.Dataset: The original dataset.

        See Also:
            :meth:`get_dataset`
        """
        if 'transform' not in kwargs.keys():
            kwargs['transform'] = self.get_transform(mode=mode)
        return self._get_org_dataset(mode=mode, **kwargs)

    @abstractmethod
    def _get_org_dataset(self, mode: str, transform: object = None,
                         **kwargs) -> torch.utils.data.Dataset:
        ...

    def get_dataset(self, mode: str = None, seed: int = None,
                    class_list: None | int | list[int] = None,
                    **kwargs):
        r"""Get dataset. Call :meth:`split_dataset` to split the training set
        if :attr:`valid_set` is ``False``.

        Args:
            mode (str): Dataset mode (e.g., ``'train'`` or ``'valid'``).
            seed (int): The random seed to split dataset
                using :any:`numpy.random.shuffle`.
                Defaults to ``env['data_seed']``.
            class_list (int | list[int]):
                The class list to pick. Defaults to ``None``.
            **kwargs: Keyword arguments passed to :meth:`get_org_dataset`.

        Returns:
            torch.utils.data.Dataset: The original dataset.
        """
        try:
            if self.valid_set or mode not in ['train', 'valid']:
                dataset = self.get_org_dataset(mode=mode, **kwargs)
            else:
                dataset = self.get_org_dataset(mode='train', **kwargs)
                subset: dict[str, torch.utils.data.Subset] = {}
                subset['train'], subset['valid'] = self.split_dataset(
                    dataset, percent=self.split_ratio, seed=seed)
                dataset = subset[mode]
        except RuntimeError:
            print(f'{self.folder_path=}')
            raise
        if class_list is not None:
            dataset = self.get_class_subset(dataset=dataset, class_list=class_list)
        return dataset

    @staticmethod
    def split_dataset(dataset: torch.utils.data.Dataset | torch.utils.data.Subset,
                      length: int = None, percent: float = None,
                      shuffle: bool = True, seed: int = None):
        r"""Split a dataset into two subsets.

        Args:
            dataset (torch.utils.data.Dataset): The dataset to split.
            length (int): The length of the first subset.
                This argument cannot be used together with :attr:`percent`.
                If ``None``, use :attr:`percent` to calculate length instead.
                Defaults to ``None``.
            percent (float): The split ratio for the first subset.
                This argument cannot be used together with :attr:`length`.
                ``length = percent * len(dataset)``.
                Defaults to ``None``.
            shuffle (bool): Whether to shuffle the dataset.
                Defaults to ``True``.
            seed (bool): The random seed to split dataset
                using :any:`numpy.random.shuffle`.
                Defaults to ``env['data_seed']``.

        Returns:
            (torch.utils.data.Subset, torch.utils.data.Subset):
                The two splitted subsets.

        :Example:
            >>> from trojanzoo.utils.data import TensorListDataset
            >>> from trojanzoo.datasets import Dataset
            >>> import torch
            >>>
            >>> data = torch.ones(11, 3, 32, 32)
            >>> targets = list(range(11))
            >>> dataset = TensorListDataset(data, targets)
            >>> set1, set2 = Dataset.split_dataset(dataset, length=3)
            >>> len(set1), len(set2)
            (3, 8)
            >>> set3, set4 = split_dataset(dataset, percent=0.5)
            >>> len(set3), len(set4)
            (5, 6)

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.data.split_dataset`.
            The difference is that this method will set :attr:`seed`
            as ``env['data_seed']`` when it is ``None``.
        """
        seed = env['data_seed'] if seed is None else seed
        return split_dataset(dataset, length=length, percent=percent,
                             shuffle=shuffle, seed=seed)

    @staticmethod
    def get_class_subset(dataset: torch.utils.data.Dataset,
                         class_list: int | list[int]) -> torch.utils.data.Subset:
        r"""Get a subset from dataset with certain classes.

        Args:
            dataset (torch.utils.data.Dataset): The entire dataset.
            class_list (int | list[int]): The class list to pick.

        Returns:
            torch.utils.data.Subset:
                The subset with labels in :attr:`class_list`.

        :Example:
            >>> from trojanzoo.utils.data import TensorListDataset
            >>> from trojanzoo.utils.data import get_class_subset
            >>> import torch
            >>>
            >>> data = torch.ones(11, 3, 32, 32)
            >>> targets = list(range(11))
            >>> dataset = TensorListDataset(data, targets)
            >>> subset = get_class_subset(dataset, class_list=[2, 3])
            >>> len(subset)
            2

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.data.get_class_subset`.
        """
        return get_class_subset(dataset=dataset, class_list=class_list)

    def get_dataloader(self, mode: str = None,
                       dataset: torch.utils.data.Dataset = None,
                       batch_size: int = None, shuffle: bool = None,
                       num_workers: int = None, pin_memory: bool = True,
                       drop_last: bool = False, collate_fn: Callable = None,
                       **kwargs) -> torch.utils.data.DataLoader:
        r"""Get dataloader. Call :meth:`get_dataset` if :attr:`dataset` is not provided.

        Args:
            mode (str): Dataset mode (e.g., ``'train'`` or ``'valid'``).
            dataset (torch.utils.data.Dataset): The pytorch dataset.
            batch_size (int):
                Defaults to :attr:`self.batch_size` for ``'train'`` mode
                and :attr:`self.valid_batch_size` for ``'valid'`` mode.
            shuffle (bool): Whether to shuffle.
                Defaults to ``True`` for ``'train'`` mode
                and ``False`` for ``'valid'`` mode.
            num_workers (int): Number of workers for dataloader.
                Defaults to :attr:`self.num_workers`.
            pin_memory (bool): Whether to use pin memory.
                Defaults to ``True`` if there is any GPU available.
            drop_last (bool): Whether drop the last batch if not full size.
                Defaults to ``False``.
            collate_fn (~collections.abc.Callable):
                Passed to :any:`torch.utils.data.DataLoader`.
            **kwargs: Keyword arguments passed to :meth:`get_dataset`
                if :attr:`dataset` is not provided.

        Returns:
            torch.utils.data.DataLoader: The pytorch dataloader.
        """
        if batch_size is None:
            match mode:
                case 'train':
                    batch_size = self.batch_size
                case 'valid':
                    batch_size = self.valid_batch_size
                case _:
                    raise ValueError(f'{mode=}')
        if shuffle is None:
            shuffle = (mode == 'train')
        if num_workers is None:
            num_workers = self.num_workers
        if dataset is None:
            dataset = self.get_dataset(mode=mode, **kwargs)
        pin_memory = pin_memory and env['num_gpus']
        collate_fn = collate_fn or self.collate_fn
        return torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, shuffle=shuffle,
            num_workers=num_workers, pin_memory=pin_memory,
            drop_last=drop_last, collate_fn=collate_fn)

    def get_loss_weights(self, file_path: str = None,
                         verbose: bool = True) -> torch.Tensor:
        r"""Calculate :attr:`loss_weights` as reciprocal of data size of each class
        (to mitigate data imbalance).

        Args:
            file_path (str):
                | The file path of saved weights file.
                | If exist, just load the file and return;
                | else, calculate the weights, save and return.
                | Defaults to ``{folder_path}/loss_weights.npy``
            verbose (bool): Whether to print verbose information.
                Defaults to ``True``.

        Returns:
            torch.Tensor: The tensor of loss weights w.r.t. each class.
        """
        file_path = file_path if file_path is not None \
            else os.path.join(self.folder_path, 'loss_weights.npy')
        if os.path.exists(file_path):
            loss_weights = np.load(file_path)
        else:
            if verbose:
                print('Calculating Loss Weights')
            dataset = self.get_dataset('train', transform=None)
            targets = np.array(list(zip(*dataset))[1])
            loss_weights = np.reciprocal(np.bincount(targets).astype(float))
            assert len(loss_weights) == self.num_classes
            np.save(file_path, loss_weights)
            if verbose:
                print('Loss Weights Saved at ', file_path)
        return torch.from_numpy(loss_weights).to(device=env['device'], dtype=torch.float)


def add_argument(parser: argparse.ArgumentParser, dataset_name: str = None,
                 dataset: str | Dataset = None,
                 config: Config = config,
                 class_dict: dict[str, type[Dataset]] = {}
                 ) -> argparse._ArgumentGroup:
    r"""
    | Add dataset arguments to argument parser.
    | For specific arguments implementation, see :meth:`Dataset.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        dataset_name (str): The dataset name.
        dataset (str | Dataset): Dataset instance or dataset name
            (as the alias of `dataset_name`).
        config (Config): The default parameter config,
            which contains the default dataset name if not provided.
        class_dict (dict[str, type[Dataset]]):
            Map from dataset name to dataset class.
            Defaults to ``{}``.
    """
    dataset_name = get_name(
        name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    dataset_name = dataset_name if dataset_name is not None \
        else config.full_config['dataset']['default_dataset']
    group = parser.add_argument_group(
        '{yellow}dataset{reset}'.format(**ansi), description=dataset_name)
    try:
        DatasetType = class_dict[dataset_name]
    except KeyError:
        print(f'{dataset_name} not in \n{list(class_dict.keys())}')
        raise
    return DatasetType.add_argument(group)


def create(dataset_name: str = None, dataset: str = None,
           config: Config = config,
           class_dict: dict[str, type[Dataset]] = {},
           **kwargs) -> Dataset:
    r"""
    | Create a dataset instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{data_dir}/{data_type}/{name}'``.
    | For dataset implementation, see :class:`Dataset`.

    Args:
        dataset_name (str): The dataset name.
        dataset (str): The alias of `dataset_name`.
        config (Config): The default parameter config.
        class_dict (dict[str, type[Dataset]]):
            Map from dataset name to dataset class.
            Defaults to ``{}``.
        **kwargs: Keyword arguments
            passed to dataset init method.

    Returns:
        Dataset: Dataset instance.
    """
    dataset_name = get_name(
        name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    dataset_name = dataset_name if dataset_name is not None \
        else config.full_config['dataset']['default_dataset']
    result = config.get_config(dataset_name=dataset_name)[
        'dataset'].update(kwargs)
    try:
        DatasetType = class_dict[dataset_name]
    except KeyError:
        print(f'{dataset_name} not in \n{list(class_dict.keys())}')
        raise
    if 'folder_path' not in result.keys():
        result['folder_path'] = os.path.join(result['data_dir'],
                                             DatasetType.data_type,
                                             DatasetType.name)
    return DatasetType(**result)



#!/usr/bin/env python3

from trojanvision.datasets.imagefolder import ImageFolder

from .cub200 import CUB200, CUB200_2011
from .gtsrb import GTSRB
from .imagenet import ImageNet, Sample_ImageNet
from .isic import ISIC2018
from .vggface2 import VGGface2, Sample_VGGface2


__all__ = ['CUB200', 'CUB200_2011', 'GTSRB', 'ImageNet', 'ISIC2018',
           'VGGface2', 'Sample_ImageNet', 'Sample_VGGface2', ]

class_dict: dict[str, ImageFolder] = {
    'cub200': CUB200,
    'cub200_2011': CUB200_2011,
    'gtsrb': GTSRB,
    'imagenet': ImageNet,
    'isic2018': ISIC2018,
    'vggface2': VGGface2,
    'sample_imagenet': Sample_ImageNet,
    'sample_vggface2': Sample_VGGface2,
}



#!/usr/bin/env python3

from trojanvision.datasets.imagefolder import ImageFolder
from trojanzoo.utils.output import ansi, prints

from torchvision.datasets.utils import download_file_from_google_drive, extract_archive, check_integrity
import os
import shutil
import pandas as pd

from trojanvision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class CUB200(ImageFolder):
    r"""CUB200 dataset introduced by Peter Welinder in 2010.
    It inherits :class:`trojanvision.datasets.ImageFolder`.

    See Also:
        * paper: `Caltech-UCSD Birds 200`_
        * website: http://www.vision.caltech.edu/visipedia/CUB-200.html

    Attributes:
        name (str): ``'cub200'``
        num_classes (int): ``200``
        data_shape (list[int]): ``[3, 224, 224]``
        valid_set (bool): ``False``

    .. _Caltech-UCSD Birds 200:
        http://www.vision.caltech.edu/visipedia/papers/WelinderEtal10_CUB-200.pdf
    """
    name = 'cub200'
    num_classes = 200
    valid_set = False
    # http://www.vision.caltech.edu/visipedia-data/CUB-200/images.tgz
    url = {'train': '1GDr1OkoXdhaXWGA8S3MAq3a522Tak-nx'}
    ext = {'train': '.tgz'}
    md5 = {'train': '2bbe304ef1aa3ddb6094aa8f53487cf2'}

    org_folder_name = {'train': 'images'}

    def download_and_extract_archive(self, mode: str):
        file_name = f'{self.name}_{mode}{self.ext[mode]}'
        file_path = os.path.normpath(os.path.join(self.folder_path, file_name))
        md5 = self.md5.get(mode)
        if not check_integrity(file_path, md5=md5):
            prints('{yellow}Downloading Dataset{reset} '.format(**ansi),
                   f'{self.name} {mode:5s}: {file_path}', indent=10)
            download_file_from_google_drive(file_id=self.url[mode],
                                            root=self.folder_path, filename=file_name, md5=md5)
            print('{upline}{clear_line}'.format(**ansi))
        else:
            prints('{yellow}File Already Exists{reset}: '.format(**ansi), file_path, indent=10)
        extract_archive(from_path=file_path, to_path=self.folder_path)

    def initialize_folder(self, **kwargs):
        super().initialize_folder(**kwargs)
        self.split()

    def split(self):
        # Remove useless files
        os.remove(os.path.join(self.folder_path, '._images'))
        dirpath = os.path.join(self.folder_path, 'train')
        for fpath in os.listdir(dirpath):
            path = os.path.join(dirpath, fpath)
            if os.path.isfile(path):
                os.remove(path)

        # Split Train and Valid Set
        txt_path = os.path.normpath(os.path.join(root_dir, 'data', 'cub200', 'test.txt'))
        file_list: list[str] = []
        with open(txt_path, 'r') as fp:
            file_list = fp.read().split('\n')[:-1]
        src_dir = os.path.join(self.folder_path, 'train')
        dst_dir = os.path.join(self.folder_path, 'valid')
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for fpath in file_list:
            src_path = os.path.join(src_dir, fpath)
            dst_path = os.path.join(dst_dir, fpath)
            dir_name = os.path.dirname(dst_path)
            if not os.path.exists(dir_name):
                os.makedirs(dir_name)
            shutil.move(src_path, dst_path)


class CUB200_2011(CUB200):
    r"""CUB200_2011 dataset. It inherits :class:`trojanvision.datasets.ImageFolder`.

    See Also:
        * paper: `The Caltech-UCSD Birds-200-2011 Dataset`_
        * website: http://www.vision.caltech.edu/visipedia/CUB-200-2011.html

    Attributes:
        name (str): ``'cub200_2011'``
        num_classes (int): ``200``
        data_shape (list[int]): ``[3, 224, 224]``
        valid_set (bool): ``False``

    .. _The Caltech-UCSD Birds-200-2011 Dataset:
        http://www.vision.caltech.edu/visipedia/papers/CUB_200_2011.pdf
    """

    name = 'cub200_2011'
    # http://www.vision.caltech.edu/visipedia-data/CUB-200-2011/CUB_200_2011.tgz
    url = {'train': '1hbzc_P1FuxMkcabkgn9ZKinBwW683j45'}
    ext = {'train': '.tgz'}
    md5 = {'train': '97eceeb196236b17998738112f37df78'}

    org_folder_name = {'train': 'CUB_200_2011/images'}

    def split(self):
        # Split Train and Valid Set
        src_dir = os.path.join(self.folder_path, 'total')
        dst_dir = {'train': os.path.join(self.folder_path, 'train'),
                   'valid': os.path.join(self.folder_path, 'valid')}
        os.rename(dst_dir['train'], src_dir)
        os.remove(os.path.join(self.folder_path, 'attributes.txt'))

        images = pd.read_csv(os.path.join(root_dir, 'data', 'cub200_2011', 'images.txt'),
                             sep=' ', names=['img_id', 'filepath'])
        train_test_split = pd.read_csv(os.path.join(root_dir, 'data', 'cub200_2011', 'train_test_split.txt'),
                                       sep=' ', names=['img_id', 'is_training_img'])
        data = images.merge(train_test_split, on='img_id')
        file_dict: dict[str, list[str]] = {
            'train': data[data.is_training_img == 1]['filepath'].tolist(),
            'valid': data[data.is_training_img == 0]['filepath'].tolist(),
        }
        for mode in ['train', 'valid']:
            for fpath in file_dict[mode]:
                src_path = os.path.join(src_dir, fpath)
                dst_path = os.path.join(dst_dir[mode], fpath)
                dir_name = os.path.dirname(dst_path)
                if not os.path.exists(dir_name):
                    os.makedirs(dir_name)
                shutil.move(src_path, dst_path)
        shutil.rmtree(src_dir)



#!/usr/bin/env python3

from trojanvision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
import torchvision.models
from torchvision.models.efficientnet import model_urls as urls

from collections.abc import Callable


class _EfficientNet(_ImageModel):

    def __init__(self, name: str = 'efficientnet_b0', **kwargs):
        super().__init__(**kwargs)
        ModelClass: Callable[..., torchvision.models.EfficientNet]
        ModelClass = getattr(torchvision.models, name.replace('_comp', ''))
        _model = ModelClass(num_classes=self.num_classes)
        self.features = _model.features
        self.classifier = _model.classifier
        # nn.Sequential(
        #     nn.Dropout(p=dropout, inplace=True),
        #     nn.Linear(lastconv_output_channels, num_classes),
        # )
        if 'comp' in name:
            conv: nn.Conv2d = self.features[0][0]
            conv = nn.Conv2d(conv.in_channels, conv.out_channels,
                             kernel_size=3, padding=1, bias=False)
            self.features[0][0] = conv


class EfficientNet(ImageModel):
    r"""EfficientNet proposed by Mingxing Tan from Google in ICML 2019.

    :Available model names:

        .. code-block:: python3

            ['efficientnet', 'efficientnet_comp',
             'efficientnet_b0', 'efficientnet_b1', 'efficientnet_b2',
             'efficientnet_b3', 'efficientnet_b4', 'efficientnet_b5',
             'efficientnet_b6', 'efficientnet_b7',
             'efficientnet_b0_comp', 'efficientnet_b1_comp', 'efficientnet_b2_comp',
             'efficientnet_b3_comp', 'efficientnet_b4_comp', 'efficientnet_b5_comp',
             'efficientnet_b6_comp', 'efficientnet_b7_comp']

    See Also:
        * torchvision: :any:`torchvision.models.efficientnet_b0`
        * paper: `EfficientNet\: Rethinking Model Scaling for Convolutional Neural Networks`_

    Note:
        ``_comp`` reduces the first convolutional layer
        from ``kernel_size=7, stride=2, padding=3``

        to ``kernel_size=3, stride=1, padding=1``.

    .. _EfficientNet\: Rethinking Model Scaling for Convolutional Neural Networks:
        https://arxiv.org/abs/1905.11946
    """
    available_models = ['efficientnet', 'efficientnet_comp',
                        'efficientnet_b0', 'efficientnet_b1', 'efficientnet_b2',
                        'efficientnet_b3', 'efficientnet_b4', 'efficientnet_b5',
                        'efficientnet_b6', 'efficientnet_b7',
                        'efficientnet_b0_comp', 'efficientnet_b1_comp', 'efficientnet_b2_comp',
                        'efficientnet_b3_comp', 'efficientnet_b4_comp', 'efficientnet_b5_comp',
                        'efficientnet_b6_comp', 'efficientnet_b7_comp']
    model_urls = urls

    def __init__(self, name: str = 'efficientnet', layer: str = '_b0',
                 model: type[_EfficientNet] = _EfficientNet, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    @classmethod
    def get_name(cls, name: str, layer: str = '') -> str:
        layer = layer if '_b' not in name else ''
        return super().get_name(name, layer=layer)



#!/usr/bin/env python3

from .imagemodel import _ImageModel, ImageModel

from .nas import *
from .normal import *
from .others import *
from .torchvision import *

from . import nas, normal, others, torchvision

from trojanvision.configs import config
import trojanzoo.models

from trojanvision.datasets import ImageSet
from trojanzoo.configs import Config
import argparse

module_list = [nas, normal, others, torchvision]
__all__ = ['_ImageModel', 'ImageModel',
           'add_argument', 'create',
           'output_available_models']
class_dict: dict[str, type[ImageModel]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)


def add_argument(parser: argparse.ArgumentParser, model_name: str = None, model: str | ImageModel = None,
                 config: Config = config, class_dict: dict[str, type[ImageModel]] = class_dict):
    r"""
    | Add image model arguments to argument parser.
    | For specific arguments implementation, see :meth:`ImageModel.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        model_name (str): The model name.
        model (str | ImageModel): Model instance or model name
            (as the alias of `model_name`).
        config (Config): The default parameter config,
            which contains the default dataset and model name if not provided.
        class_dict (dict[str, type[ImageModel]]):
            Map from model name to model class.
            Defaults to ``trojanvision.models.class_dict``.

    See Also:
        :func:`trojanzoo.models.add_argument()`
    """
    return trojanzoo.models.add_argument(parser=parser, model_name=model_name, model=model,
                                         config=config, class_dict=class_dict)


def create(model_name: str = None, model: str | ImageModel = None,
           dataset_name: str = None, dataset: str | ImageSet = None,
           config: Config = config, class_dict: dict[str, type[ImageModel]] = class_dict,
           **kwargs) -> ImageModel:
    r"""
    | Create a model instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{model_dir}/{dataset.data_type}/{dataset.name}'``.
    | For model implementation, see :class:`ImageModel`.

    Args:
        model_name (str): The model name.
        model (str | ImageModel): The model instance or model name
            (as the alias of `model_name`).
        dataset_name (str): The dataset name.
        dataset (str | trojanvision.datasets.ImageSet):
            Dataset instance or dataset name
            (as the alias of `dataset_name`).
        config (Config): The default parameter config.
        class_dict (dict[str, type[ImageModel]]):
            Map from model name to model class.
            Defaults to ``trojanvision.models.class_dict``.
        **kwargs: Keyword arguments
            passed to model init method.

    Returns:
        ImageModel: The image model instance.

    See Also:
        :func:`trojanzoo.models.create()`
    """
    return trojanzoo.models.create(model_name=model_name, model=model,
                                   dataset_name=dataset_name, dataset=dataset,
                                   config=config, class_dict=class_dict, **kwargs)


def output_available_models(class_dict: dict[str, type[ImageModel]] = class_dict, indent: int = 0) -> None:
    r"""Output all available model names.

    Args:
        class_dict (dict[str, type[ImageModel]]): Map from model name to model class.
            Defaults to ``trojanvision.models.class_dict``.
        indent (int): The space indent for the entire string.
            Defaults to ``0``.

    See Also:
        :func:`trojanzoo.models.output_available_models()`
    """
    return trojanzoo.models.output_available_models(class_dict=class_dict, indent=indent)


def get_available_models(class_dict: dict[str, type[ImageModel]] = class_dict) -> dict[str, list[str]]:
    return trojanzoo.models.get_available_models(class_dict=class_dict)



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --epochs 10 --lr 0.01 --mark_random_init --attack latent_backdoor
"""  # noqa: E501

from ...abstract import BackdoorAttack

from trojanvision.environ import env
from trojanzoo.utils.data import sample_batch
from trojanzoo.utils.tensor import tanh_func

import torch
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import TensorDataset
import functools

from typing import TYPE_CHECKING
import argparse
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data


class LatentBackdoor(BackdoorAttack):
    r"""Latent Backdoor proposed by Yuanshun Yao, Huiying Li, Haitao Zheng
    and Ben Y. Zhao from University of Chicago in CCS 2019.

    It inherits :class:`trojanvision.attacks.BackdoorAttack`.

    Similar to :class:`trojanvision.attacks.TrojanNN`,
    Latent Backdoor preprocesses watermark pixel values to
    minimize feature mse distance (of other classes with trigger attached)
    to average feature map of target class.

    Loss formulas are:

    * ``'preprocess'``: :math:`\mathcal{L}_{MSE}`
    * ``'retrain'``: :math:`\mathcal{L}_{CE} + \text{self.mse\_weight} * \mathcal{L}_{MSE}`

    See Also:
        * paper: `Latent Backdoor Attacks on Deep Neural Networks`_
        * code: https://github.com/Huiying-Li/Latent-Backdoor
        * website: https://sandlab.cs.uchicago.edu/latent

    Note:
        This implementation does **NOT** involve
        teacher-student transfer learning nor new learning tasks,
        which are main contribution and application scenario of the original paper.
        It still focuses on BadNet problem setting and
        only utilizes the watermark optimization and retraining loss from Latent Backdoor attack.

        For users who have those demands, please inherit this class and use the methods as utilities.

    Args:
        class_sample_num (int): Sampled input number of each class.
            Defaults to ``100``.
        mse_weight (float): MSE loss weight used in model retraining.
            Defaults to ``0.5``.
        preprocess_layer (str): The chosen layer to calculate feature map.
            Defaults to ``'flatten'``.
        attack_remask_epochs (int): Watermark preprocess optimization epoch.
            Defaults to ``100``.
        attack_remask_lr (float): Watermark preprocess optimization learning rate.
            Defaults to ``0.1``.

    .. _Latent Backdoor Attacks on Deep Neural Networks:
        https://dl.acm.org/doi/10.1145/3319535.3354209
    """
    name: str = 'latent_backdoor'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--class_sample_num', type=int,
                           help='sampled input number of each class '
                           '(default: 100)')
        group.add_argument('--mse_weight', type=float,
                           help='MSE loss weight used in model retraining '
                           '(default: 0.5)')
        group.add_argument('--preprocess_layer',
                           help='the chosen layer to calculate feature map '
                           '(default: "flatten")')
        group.add_argument('--attack_remask_epochs', type=int,
                           help='preprocess optimization epochs')
        group.add_argument('--attack_remask_lr', type=float,
                           help='preprocess learning rate')
        return group

    def __init__(self, class_sample_num: int = 100, mse_weight: float = 0.5,
                 preprocess_layer: str = 'flatten',
                 attack_remask_epochs: int = 100, attack_remask_lr: float = 0.1,
                 **kwargs):
        super().__init__(**kwargs)
        if not self.mark.mark_random_init:
            raise Exception('Latent Backdoor requires "mark_random_init" to be True to initialize watermark.')
        if self.mark.mark_random_pos:
            raise Exception('Latent Backdoor requires "mark_random_pos" to be False.')

        self.param_list['latent_backdoor'] = ['class_sample_num', 'mse_weight',
                                              'preprocess_layer', 'attack_remask_epochs', 'attack_remask_lr']
        self.class_sample_num = class_sample_num
        self.mse_weight = mse_weight

        self.preprocess_layer = preprocess_layer
        self.attack_remask_epochs = attack_remask_epochs
        self.attack_remask_lr = attack_remask_lr

        self.avg_target_feats: torch.Tensor = None

    def attack(self, **kwargs):
        print('Sample Data')
        data = self.sample_data()
        print('Calculate Average Target Features')
        self.avg_target_feats = self.get_avg_target_feats(*data['target'])
        print('Preprocess Mark')
        self.preprocess_mark(*data['other'])
        print('Retrain')
        if 'loss_fn' in kwargs.keys():
            kwargs['loss_fn'] = functools.partial(self.loss, loss_fn=kwargs['loss_fn'])
        else:
            kwargs['loss_fn'] = self.loss
        return super().attack(**kwargs)

    def sample_data(self) -> dict[str, tuple[torch.Tensor, torch.Tensor]]:
        r"""Sample data from each class. The returned data dict is:

        * ``'other'``: ``(input, label)`` from source classes with batch size
          ``self.class_sample_num * len(source_class)``.
        * ``'target'``: ``(input, label)`` from target class with batch size
          ``self.class_sample_num``.

        Returns:
            dict[str, tuple[torch.Tensor, torch.Tensor]]: Data dict.
        """
        source_class = self.source_class or list(range(self.dataset.num_classes))
        source_class = source_class.copy()
        if self.target_class in source_class:
            source_class.remove(self.target_class)
        other_x, other_y = [], []
        dataset = self.dataset.get_dataset('train')
        for _class in source_class:
            class_set = self.dataset.get_class_subset(dataset, class_list=[_class])
            _input, _label = sample_batch(class_set, batch_size=self.class_sample_num)
            other_x.append(_input)
            other_y.append(_label)
        other_x = torch.cat(other_x)
        other_y = torch.cat(other_y)
        target_set = self.dataset.get_class_subset(dataset, class_list=[self.target_class])
        target_x, target_y = sample_batch(target_set, batch_size=self.class_sample_num)
        data = {'other': (other_x, other_y),
                'target': (target_x, target_y)}
        return data

    @torch.no_grad()
    def get_avg_target_feats(self, target_input: torch.Tensor,
                             target_label: torch.Tensor
                             ) -> torch.Tensor:
        r"""Get average feature map of :attr:`self.preprocess_layer`
        using sampled data from :attr:`self.target_class`.

        Args:
            target_input (torch.Tensor): Input tensor from target class with shape
                ``(self.class_sample_num, C, H, W)``.
            target_label (torch.Tensor): Label tensor from target class with shape
                ``(self.class_sample_num)``.

        Returns:
            torch.Tensor:
                Feature map tensor with shape
                ``(self.class_sample_num, C')``.
        """
        if self.dataset.data_shape[1] > 100:
            dataset = TensorDataset(target_input, target_label)
            loader = torch.utils.data.DataLoader(
                dataset=dataset, batch_size=self.dataset.batch_size // max(env['num_gpus'], 1),
                num_workers=0, pin_memory=True)
            feat_list = []
            for data in loader:
                target_x, _ = self.model.get_data(data)
                feat_list.append(self.model.get_layer(
                    target_x, layer_output=self.preprocess_layer).detach().cpu())
            avg_target_feats = torch.cat(feat_list).mean(dim=0, keepdim=True)
            avg_target_feats = avg_target_feats.to(target_x.device)
        else:
            target_input, _ = self.model.get_data((target_input, target_label))
            avg_target_feats = self.model.get_layer(
                target_input, layer_output=self.preprocess_layer).mean(dim=0, keepdim=True)
        if avg_target_feats.dim() > 2:
            avg_target_feats = avg_target_feats.flatten(2).mean(2)
        return avg_target_feats.detach()

    def preprocess_mark(self, other_input: torch.Tensor, other_label: torch.Tensor):
        r"""Preprocess to optimize watermark using data sampled from source classes.

        Args:
            other_input (torch.Tensor): Input tensor from source classes with shape
                ``(self.class_sample_num * len(source_class), C, H, W)``.
            other_label (torch.Tensor): Label tensor from source classes with shape
                ``(self.class_sample_num * len(source_class))``.
        """
        other_set = TensorDataset(other_input, other_label)
        other_loader = self.dataset.get_dataloader(mode='train', dataset=other_set, num_workers=0)

        atanh_mark = torch.randn_like(self.mark.mark[:-1], requires_grad=True)
        optimizer = optim.Adam([atanh_mark], lr=self.attack_remask_lr)
        optimizer.zero_grad()

        for _ in range(self.attack_remask_epochs):
            for data in other_loader:
                self.mark.mark[:-1] = tanh_func(atanh_mark)
                _input, _label = self.model.get_data(data)
                trigger_input = self.add_mark(_input)
                loss = self._loss_mse(trigger_input)
                loss.backward(inputs=[atanh_mark])
                optimizer.step()
                optimizer.zero_grad()
                self.mark.mark.detach_()
        atanh_mark.requires_grad_(False)
        self.mark.mark[:-1] = tanh_func(atanh_mark)
        self.mark.mark.detach_()

    # -------------------------------- Loss Utils ------------------------------ #
    def loss(self, _input: torch.Tensor, _label: torch.Tensor,
             loss_fn: Callable[..., torch.Tensor] = None,
             **kwargs) -> torch.Tensor:
        loss_fn = loss_fn if loss_fn is not None else self.model.loss
        loss_ce = loss_fn(_input, _label, **kwargs)
        trigger_input = self.add_mark(_input)
        loss_mse = self._loss_mse(trigger_input)
        return loss_ce + self.mse_weight * loss_mse

    def _loss_mse(self, trigger_input: torch.Tensor) -> torch.Tensor:
        poison_feats = self.model.get_layer(trigger_input, layer_output=self.preprocess_layer)
        if poison_feats.dim() > 2:
            poison_feats = poison_feats.flatten(2).mean(2)
        return F.mse_loss(poison_feats, self.avg_target_feats.expand(poison_feats.size(0), -1))



#!/usr/bin/env python3

from trojanzoo.utils.output import ansi, output_iter, prints, redirect

import os
import functools
from collections.abc import Iterable

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from trojanzoo.datasets import Dataset    # TODO: python 3.10
    from trojanzoo.models import Model

__all__ = ['BasicObject', 'Process', 'ModelProcess']


class BasicObject:
    r"""A basic class with a pretty :meth:`summary()` method.

    Attributes:
        name (str): The name of the instance or class.
        param_list (dict[str, list[str]]): Map from category strings to variable name list.
        indent (int): The indent when calling :meth:`summary()`. Defaults to ``0``.
    """
    name: str = 'basic_object'

    def __init__(self, indent: int = 0, **kwargs):
        self.param_list: dict[str, list[str]] = {}
        self.param_list['verbose'] = ['indent']
        self.indent = indent

    # -----------------------------------Output-------------------------------------#
    def summary(self, indent: int = None):
        r"""Summary the variables of the instance
        according to :attr:`param_list`.

        Args:
            indent (int): The space indent for the entire string.
                Defaults to :attr:`self.indent`.

        See Also:
            :meth:`trojanzoo.models.Model.summary()`.
        """
        indent = indent if indent is not None else self.indent
        prints('{blue_light}{0:<30s}{reset} Parameters: '.format(
            self.name, **ansi), indent=indent)
        prints('{yellow}{0}{reset}'.format(self.__class__.__name__, **ansi), indent=indent)
        for key, value in self.param_list.items():
            if value:
                prints('{green}{0:<20s}{reset}'.format(
                    key, **ansi), indent=indent + 10)
                prints({v: str(getattr(self, v)).split('\n')[0]
                       for v in value}, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)

    def __str__(self) -> str:
        with redirect():
            self.summary()
            return redirect.buffer


class Process(BasicObject):
    r"""It inherits :class:`BasicObject` and further specify output levels.

    Args:
        output (int, ~collections.abc.Iterable[str]):
            The level of output or the set of output items.

    Attributes:
        output (set[str]): The set of output items
            generated by :meth:`get_output()`.

    See Also:
        :class:`trojanzoo.optim.Optimizer`
        and :class:`ModelProcess` inherit this class.
    """
    name: str = 'process'

    def __init__(self, output: int | Iterable[str] = 0,
                 **kwargs):
        super().__init__(**kwargs)
        self.param_list['verbose'] = ['output', 'indent']

        self.output: set[str] = None
        self.output = self.get_output(output)

    # -----------------------------------Output-------------------------------------#

    def get_output(self, org_output: int | Iterable[str] = None
                   ) -> set[str]:
        r"""Get output items based on output level.

        Args:
            org_output (int, ~collections.abc.Iterable[str]):
                Output level integer or output items.
                If :class:`int`, call :meth:`get_output_int()`.
                Defaults to :attr:`self.output`.

        Returns:
            set[str]: The set of output items.
        """
        match org_output:
            case None:
                return self.output
            case int():
                return self.get_output_int(org_output)
            case _:
                return set(org_output)

    @classmethod
    def get_output_int(cls, org_output: int = 0) -> set[str]:
        r"""Get output items based on output level integer.

            * ``0  - 4 : {'verbose'}``
            * ``5  - 9 : {'verbose', 'end'}``
            * ``10 - 19: {'verbose', 'end', 'start'}``
            * ``20 - 29: {'verbose', 'end', 'start', 'middle'}``
            * ``30 - * : {'verbose', 'end', 'start', 'middle', 'memory'}``

        Args:
            org_output (int): Output level integer.
                Defaults to ``0``.

        Returns:
            set[str]: The set of output items.
        """
        result: set[str] = set()
        if org_output > 0:
            result.add('verbose')
        if org_output >= 5:
            result.add('end')
        if org_output >= 10:
            result.add('start')
        if org_output >= 20:
            result.add('middle')
        if org_output >= 30:
            result.add('memory')
        return result

    @staticmethod
    def output_iter(name: str, _iter: int,
                    iteration: int = None) -> str:
        r"""Output an iteration string:
        ``{name} Iter: [ {_iter + 1} / {iteration} ]``
        or ``{name} Iter: [ {_iter + 1} ]``
        if :attr:`iteration` is ``None``.

        Args:
            name (str): The header string.
            _iter (int): The current iteration.
            iteration (int): The total iteration.
                Defaults to ``None``.

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.output.output_iter()`.
        """
        return f'{name} Iter: {output_iter(_iter + 1, iteration)}'


class ModelProcess(Process):
    r"""It inherits :class:`Process`
    and further specify model related items.

    Attributes:
        dataset (trojanzoo.datasets.Dataset | None): The dataset instance.
        model (trojanzoo.models.Model | None): The model instance.
        folder_path (str | None): The folder path to store results.
            Defaults to ``None``.
        clean_acc (float): The clean accuracy of :attr:`model`.

    See Also:
        :class:`trojanzoo.attacks.Attack`
        and :class:`trojanzoo.defenses.Defense`
        inherit this class.
    """
    name: str = 'ModelProcess'

    def __init__(self, dataset: 'Dataset' = None, model: 'Model' = None,
                 folder_path: str = None, **kwargs):
        super().__init__(**kwargs)
        self.param_list['process'] = ['folder_path', 'clean_acc']
        self.dataset = dataset
        self.model = model

        if folder_path is not None:
            folder_path = os.path.normpath(folder_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
        self.folder_path = folder_path

    @functools.cached_property
    def clean_acc(self) -> float:
        clean_acc, _ = self.model._validate(verbose=False)
        return clean_acc



#!/usr/bin/env python3

# https://github.com/google-research/big_transfer

r"""
CUDA_VISIBLE_DEVICES=0 python examples/validate.py --color --verbose 1 --dataset imagenet --model bit-m-r50x1 --official --transform bit
"""  # noqa: E501

from trojanvision.utils.model_archs.bit import KNOWN_MODELS, tf2th, StdConv2d, conv3x3
from trojanvision.models.imagemodel import _ImageModel, ImageModel
from trojanvision.datasets import ImageNet

import torch
import torch.nn as nn
import torch.hub
import numpy as np
import os
from collections import OrderedDict


class _BiT(_ImageModel):
    def __init__(self, name: str = 'bit-m-r50x1', **kwargs):
        model_name = name.split('_')[0].upper().replace('BIT', 'BiT').replace('X', 'x')
        _model = KNOWN_MODELS[model_name](head_size=1)
        root, head = _model.root, _model.head
        if 'num_features' not in kwargs.keys():
            head_conv: nn.Conv2d = getattr(head, 'conv')
            kwargs['num_features'] = [head_conv.in_channels]
        super().__init__(**kwargs)

        self.features = nn.Sequential()
        root_conv: StdConv2d = getattr(root, 'conv')
        if 'comp' in name:
            self.features.add_module('conv', conv3x3(root_conv.in_channels, root_conv.out_channels))
        else:
            self.features.add_module('conv', root_conv)
            root_pool: nn.MaxPool2d = getattr(root, 'pool')
            if 'official' not in name:
                root_pool = nn.MaxPool2d(root_pool.kernel_size, root_pool.stride, padding=1)
            else:
                self.features.add_module('pad', getattr(root, 'pad'))
            self.features.add_module('pool', root_pool)
        for name, child in _model.body.named_children():
            self.features.add_module(name, child)
        self.features.add_module('gn', getattr(head, 'gn'))
        self.features.add_module('relu', getattr(head, 'relu'))


class BiT(ImageModel):
    r"""Big Transfer (ResNetv2) proposed by Alexander Kolesnikov from Google in ECCV 2020.

    :Available model names:

        .. code-block:: python3

            ['bit', 'bit_comp', 'bit_official',
             'bit-m-r50x1', 'bit-m-r50x3', 'bit-m-r101x1', 'bit-m-r101x3', 'bit-m-r152x2', 'bit-m-r152x4',
             'bit-s-r50x1', 'bit-s-r50x3', 'bit-s-r101x1', 'bit-s-r101x3', 'bit-s-r152x2', 'bit-s-r152x4',
             'bit-m-r50x1_comp', 'bit-m-r50x3_comp', 'bit-m-r101x1_comp',
             'bit-m-r101x3_comp', 'bit-m-r152x2_comp', 'bit-m-r152x4_comp',
             'bit-s-r50x1_comp', 'bit-s-r50x3_comp', 'bit-s-r101x1_comp',
             'bit-s-r101x3_comp', 'bit-s-r152x2_comp', 'bit-s-r152x4_comp',
             'bit-m-r50x1_official', 'bit-m-r50x3_official', 'bit-m-r101x1_official',
             'bit-m-r101x3_official', 'bit-m-r152x2_official', 'bit-m-r152x4_official',
             'bit-s-r50x1_official', 'bit-s-r50x3_official', 'bit-s-r101x1_official',
             'bit-s-r101x3_official', 'bit-s-r152x2_official', 'bit-s-r152x4_official']

    See Also:
        * paper: `Big Transfer (BiT)\: General Visual Representation Learning`_
        * code: https://github.com/google-research/big_transfer

    Note:
        ``_comp`` reduces the first convolutional layer
        from ``kernel_size=7, stride=2, padding=3``

        to ``kernel_size=3, stride=1, padding=1``,
        and removes following ``norm0, relu0, pool0``
        (``pool0`` is :any:`torch.nn.MaxPool2d`)
        before block layers.

    .. _Big Transfer (BiT)\: General Visual Representation Learning:
        https://arxiv.org/abs/1912.11370
    """
    available_models = ['bit', 'bit_comp', 'bit_official',
                        'bit-m-r50x1', 'bit-m-r50x3', 'bit-m-r101x1', 'bit-m-r101x3', 'bit-m-r152x2', 'bit-m-r152x4',
                        'bit-s-r50x1', 'bit-s-r50x3', 'bit-s-r101x1', 'bit-s-r101x3', 'bit-s-r152x2', 'bit-s-r152x4',
                        'bit-m-r50x1_comp', 'bit-m-r50x3_comp', 'bit-m-r101x1_comp',
                        'bit-m-r101x3_comp', 'bit-m-r152x2_comp', 'bit-m-r152x4_comp',
                        'bit-s-r50x1_comp', 'bit-s-r50x3_comp', 'bit-s-r101x1_comp',
                        'bit-s-r101x3_comp', 'bit-s-r152x2_comp', 'bit-s-r152x4_comp',
                        'bit-m-r50x1_official', 'bit-m-r50x3_official', 'bit-m-r101x1_official',
                        'bit-m-r101x3_official', 'bit-m-r152x2_official', 'bit-m-r152x4_official',
                        'bit-s-r50x1_official', 'bit-s-r50x3_official', 'bit-s-r101x1_official',
                        'bit-s-r101x3_official', 'bit-s-r152x2_official', 'bit-s-r152x4_official']

    def __init__(self, name: str = 'bit',
                 pretrained_dataset: str = 'm', layer: int = 50, width_factor: int = 1,
                 model: type[_BiT] = _BiT, norm_par: dict[str, list[float]] = None,
                 official: bool = False, **kwargs):
        name = self.parse_name(name, pretrained_dataset, layer, width_factor)
        if official and norm_par is None:
            norm_par = {'mean': [0.5, 0.5, 0.5],
                        'std': [0.5, 0.5, 0.5], }
        super().__init__(name=name, width_factor=width_factor,
                         model=model, norm_par=norm_par,
                         official=official, **kwargs)

    @staticmethod
    def parse_name(name: str, pretrained_dataset: str = 'm', layer: int = 50, width_factor: int = 1) -> str:
        full_name_list = name.split('_')
        name_list = full_name_list[0].lower().split('-')
        assert name_list[0] == 'bit', name
        if len(name_list) != 1:
            for element in name_list[1:]:
                if element[0] == 'r':
                    sub_list = element[1:].split('x')
                    layer = int(sub_list[0])
                    if len(sub_list) == 2:
                        width_factor = int(sub_list[1])
                else:
                    assert len(element) == 1, name
                    pretrained_dataset = element
        full_name_list[0] = '-'.join(['bit', pretrained_dataset, f'r{layer:d}x{width_factor:d}'])
        return '_'.join(full_name_list)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        # TODO: map_location argument
        assert 'comp' not in self.name, self.name
        model_name = self.name.split('_')[0].upper().replace('BIT', 'BiT').replace('X', 'x')
        if isinstance(self.dataset, ImageNet):
            model_name += '-ILSVRC2012'
        url = f'https://storage.googleapis.com/bit_models/{model_name}.npz'
        print('get official model weights from: ', url)
        dir_path = os.path.join(torch.hub.get_dir(), 'bit')
        file_path = os.path.join(dir_path, f'{model_name}.npz')
        if not os.path.isfile(file_path):
            if not os.path.isdir(dir_path):
                os.makedirs(dir_path)
            torch.hub.download_url_to_file(url, file_path)
        weights: dict[str, np.ndarray] = np.load(file_path)
        _dict = OrderedDict()
        _dict['features.conv.weight'] = tf2th(weights['resnet/root_block/standardized_conv2d/kernel'])
        for block_num in range(4):
            block_name = f'block{block_num+1:d}'
            block: nn.Sequential = getattr(self._model.features, block_name)
            for unit_name, unit in block.named_children():
                prefix = '/'.join(['resnet', block_name, unit_name, ''])
                dict_prefix = '.'.join(['features', block_name, unit_name, ''])
                _dict[dict_prefix + 'gn1.weight'] = tf2th(weights[prefix + 'a/group_norm/gamma'])
                _dict[dict_prefix + 'gn1.bias'] = tf2th(weights[prefix + 'a/group_norm/beta'])
                _dict[dict_prefix + 'conv1.weight'] = tf2th(weights[prefix + 'a/standardized_conv2d/kernel'])
                _dict[dict_prefix + 'gn2.weight'] = tf2th(weights[prefix + 'b/group_norm/gamma'])
                _dict[dict_prefix + 'gn2.bias'] = tf2th(weights[prefix + 'b/group_norm/beta'])
                _dict[dict_prefix + 'conv2.weight'] = tf2th(weights[prefix + 'b/standardized_conv2d/kernel'])
                _dict[dict_prefix + 'gn3.weight'] = tf2th(weights[prefix + 'c/group_norm/gamma'])
                _dict[dict_prefix + 'gn3.bias'] = tf2th(weights[prefix + 'c/group_norm/beta'])
                _dict[dict_prefix + 'conv3.weight'] = tf2th(weights[prefix + 'c/standardized_conv2d/kernel'])
                if hasattr(unit, 'downsample'):
                    weight = tf2th(weights[prefix + 'a/proj/standardized_conv2d/kernel'])
                    _dict[dict_prefix + 'downsample.weight'] = weight
        _dict['features.gn.weight'] = tf2th(weights['resnet/group_norm/gamma'])
        _dict['features.gn.bias'] = tf2th(weights['resnet/group_norm/beta'])
        _dict['classifier.fc.weight'] = tf2th(weights['resnet/head/conv2d/kernel']).flatten(1)
        _dict['classifier.fc.bias'] = tf2th(weights['resnet/head/conv2d/bias'])
        return _dict

    def parametrize_(self, parametrize: bool = True):
        for mod in self.modules():
            if isinstance(mod, StdConv2d):
                mod.parametrize_(parametrize)
        return self

    def load(self, *args, **kwargs) -> OrderedDict[str, torch.Tensor]:
        self.parametrize_(False)
        _dict = super().load(*args, **kwargs)
        self.parametrize_()
        return _dict

    def save(self, *args, **kwargs):
        self.parametrize_(False)
        super().save(*args, **kwargs)
        self.parametrize_()



#!/usr/bin/env python3

from trojanzoo.attacks import Attack

from .imc_poison import IMC_Poison
from .poison_basic import PoisonBasic
from .poison_random import PoisonRandom

__all__ = ['IMC_Poison', 'PoisonBasic', 'PoisonRandom']

class_dict: dict[str, type[Attack]] = {
    'imc_poison': IMC_Poison,
    'poison_basic': PoisonBasic,
    'poison_random': PoisonRandom,
}
